
#include "Interpolator.h"
#include "Sort.h"


static void resolve(vec<Lit>& main, vec<Lit>& other, Var x);
static void printVarSet (char* name, unordered_set<Var> varSet);


int Interpolator::isVariableOfB(int literal)
{
	// ASSUMPTION! If literal is not in A, then it is in B (shared or unshared)
    return !varOnlyBelongsToA(var(toLit(literal)) + 1);
}

void Interpolator::computeVariableLists()
{
  //assert(variablesOfA.size() > 0 && variablesOfB.size() > 0);
  assert(variablesOfOnlyA.size() > 0 || variablesOfOnlyB.size() > 0);

  bool setAisSmallest = variablesOfOnlyA.size() < variablesOfOnlyB.size();
  unordered_set<Var> &smallest = setAisSmallest ? 	variablesOfOnlyA :
		  	  	  	  	  	  	  	  	  	  	  variablesOfOnlyB;
  unordered_set<Var> &largest = setAisSmallest ? 	variablesOfOnlyB :
		  	  	  	  	  	  	  	  	  	  	  variablesOfOnlyA;

  unordered_set<Var>::iterator it = smallest.begin();

  while (it != smallest.end()) {
	  Var v = *it;

	  if (largest.erase(v) > 0) {
		  variablesOfBothAB.insert(v);
		  it = smallest.erase(it);
	  } else {
		  it++;
	  }
  }

  // debug
  //printVariableLists();
  //printf("Size Only A %d\n", (int)variablesOfOnlyA.size());
  //printf("Size Only B %d\n", (int)variablesOfOnlyB.size());
  //printf("Size shared %d\n", (int)variablesOfBothAB.size());
}


void Interpolator::printVariableLists ()
{
  printf("variable lists (in minisat literals encoding)\n");

  //printVarSet("variablesOfA", variablesOfA);
  //printVarSet("variablesOfB", variablesOfB);

  printVarSet("variablesOfOnlyA", variablesOfOnlyA);
  printVarSet("variablesOfOnlyB", variablesOfOnlyB);

  printVarSet("variablesOfBothAB", variablesOfBothAB);
}

static void printSingleClause (vec<Lit>& c, int id)
{
  printf("clause %d: ", id);
  for (int j = 0; j < c.size(); j++) {
    Lit& curLit = c[j];
    int curVar = var(curLit) + 1;
    printf(" %d", sign(curLit) ? -curVar : curVar);
  }
  
  printf("\n");
}

static void printClauses (vec<vec<Lit> >& clauses)
{
  printf("printing clause base\n");

  for (int id = 0; id < clauses.size(); id++) {
    printSingleClause(clauses[id], id);
  }

  printf("\n");
}

void Interpolator::root (const vec<Lit>& c)
{
  /* store interpolation information */

  struct ItpNode * itpn = new ItpNode();

  itpn->nType = newRootsAreOfA ? ROOT_A : ROOT_B;
  itpn->pivotVariable = -1;
  itpn->interpolant = (void*)NULL;
  itpn->parentNegative = (struct ItpNode *)NULL;
  itpn->parentPositive = (struct ItpNode *)NULL;
  itpn->k = k;
  itpn->clause = (int*)malloc(c.size() * sizeof(int));
  itpn->clause_size = c.size();

  itpNodes.insert(pair<int, struct ItpNode *>(clauses.size(), itpn));
  lastItpNode = itpn;

  /* print proof step */

  if (verboseProofLogging) {
    printf("%d: ROOT (%c)", clauses.size(), newRootsAreOfA?'A':'B');
  }

  for (int i = 0; i < c.size(); i++) {
    if (verboseProofLogging) {
      printf(" %s%d", sign(c[i])?"-":"", var(c[i])+1);
    }

    /* update minisatLiteral sets */
    Var curVar = var(c[i])+1;
    if (newRootsAreOfA) {
    	addVariableOfA(curVar);
    	variableOfAi[curVar] = k;
    } else {
    	addVariableOfB(curVar);
    }

    itpn->clause[i] = index(c[i]);
  }

  if (verboseProofLogging) printf("\n");

  /* store clause */

  clauses.push();
  c.copyTo(clauses.last());
}

void Interpolator::chain (const vec<ClauseId>& cs, const vec<Var>& xs)
{
  /* print proof step (antecedent) */

  if (verboseProofLogging) {
    printf("%d: CHAIN %d", clauses.size(), cs[0]);
    for (int i = 0; i < xs.size(); i++) {
      printf(" [%d]", xs[i]+1);
      printf(" %d", cs[i+1]);
    }
  }

  /* store interpolation information */

  struct ItpNode * itpn;
  struct ItpNode * parentLeft;
  struct ItpNode * parentRight;

  for (int i = 0; i < xs.size(); i++) {

    itpn = new ItpNode();
    itpn->interpolant = (void*)NULL;
    itpn->pivotVariable = xs[i]+1;
    itpn->parentNegative = (struct ItpNode *)NULL;
    itpn->parentPositive = (struct ItpNode *)NULL;
    itpn->clause = (int*)NULL;
    itpn->clause_size = 0;
    // track in which A_i the pivot variable occurs.
    itpn->k = variableOfAi[itpn->pivotVariable];

    if (varOnlyBelongsToA(itpn->pivotVariable)) {
      itpn->nType = PIVOT_A;
    }
    else if (varOnlyBelongsToB(itpn->pivotVariable)) {
      itpn->nType = PIVOT_B;
    }
    else {
      assert(varBelongsToBothAandB(itpn->pivotVariable));
      itpn->nType = PIVOT_AB;
    }

    parentLeft = 0 == i ? itpNodes[cs[0]] : parentLeft;
    parentRight = itpNodes[cs[i+1]];

    /* identify sign of pivot variable in right parent */

    vec<Lit>& clauseParentRight = clauses[cs[i+1]];
    bool foundLiteral = false;

    for (int j = 0; j < clauseParentRight.size() && !foundLiteral; j++) {

      Lit curLit = clauseParentRight[j];
      int curVar = var(curLit) + 1;

      if (curVar == itpn->pivotVariable) {
        if (sign(curLit)) {
          itpn->parentNegative = parentRight;
          itpn->parentPositive = parentLeft;
        }
        else {
          itpn->parentNegative = parentLeft;
          itpn->parentPositive = parentRight;
        }
        foundLiteral = true;
      }
    }

    assert(foundLiteral);

    parentLeft = itpn;

    /* register anonymous nodes (needed for clean-up) */
    if (i < xs.size()-1) {
      anonymousNodes.insert(itpn);
    }
  }

  itpNodes.insert(pair<int, struct ItpNode *>(clauses.size(), itpn));
  lastItpNode = itpn;

  /* store clause */

  clauses.push();
  vec<Lit>& c = clauses.last();
  clauses[cs[0]].copyTo(c);
  for (int i = 0; i < xs.size(); i++) {
    resolve(c, clauses[cs[i+1]], xs[i]);
  }

  /* print proof step (resolvent) */

  if (verboseProofLogging) {
    printf(" =>");
    for (int i = 0; i < c.size(); i++) {
      printf(" %s%d", sign(c[i])?"-":"", var(c[i])+1);
    }
    printf("\n");
  }
}

void Interpolator::deleted (ClauseId c)
{
  clauses[c].clear();
}

/* ---------------------------------------------------------------------- */
/* Private Methods                                                        */
/* ---------------------------------------------------------------------- */

static void printVarSet (char* name, unordered_set<Var> varSet)
{
  printf("%s: ", name);

  unordered_set<Var>::iterator it;
  for (it = varSet.begin(); it != varSet.end(); it++) {
    printf("%d ", *it);
  }
  printf("\n");
}

static void resolve(vec<Lit>& main, vec<Lit>& other, Var x)
{
    Lit     p;
    bool    ok1 = false, ok2 = false;
    for (int i = 0; i < main.size(); i++){
        if (var(main[i]) == x){
            ok1 = true, p = main[i];
            main[i] = main.last();
            main.pop();
            break;
        }
    }

    for (int i = 0; i < other.size(); i++){
        if (var(other[i]) != x)
            main.push(other[i]);
        else{
            if (p != ~other[i])
                printf("PROOF ERROR! Resolved on variable with SAME polarity in both clauses: %d\n", x+1);
            ok2 = true;
        }
    }

    if (!ok1 || !ok2)
        printf("PROOF ERROR! Resolved on missing variable: %d\n", x+1);

    assert(ok1 && ok2);

    sortUnique(main);
}

