#include "SolverTypes.h"
#include "Solver.h"
#include "Proof.h"

extern "C" {
#include "Solver_C.h"
}

extern "C" MiniSat_ptr MiniSat_Create()
{
  return (MiniSat_ptr)new Solver();
}

extern "C" void MiniSat_Delete(MiniSat_ptr ms)
{
  Solver * sms = (Solver *)ms;

  if (MiniSat_InterpolationIsEnabled(ms)) {
    delete sms->proof;
    delete sms->interpolator;
  }

  delete sms;
}

extern "C" int MiniSat_Nof_Variables(MiniSat_ptr ms)
{
  return ((Solver *)ms)->nVars();
}

extern "C" int MiniSat_Nof_Clauses(MiniSat_ptr ms)
{
  return ((Solver *)ms)->nClauses();
}

/* variables are in the range 1...N */
extern "C" int MiniSat_New_Variable(MiniSat_ptr ms)
{
  /* Actually, minisat used variable range 0 .. N-1,
     so in all function below there is a convertion between
     input variable (1..N) and internal variables (0..N-1)
  */	
  return ((Solver *)ms)->newVar() + 1;
}


/*
 * Here clauses are in dimacs form, variable indexing is 1...N
 */
extern "C" int MiniSat_Add_Clause(MiniSat_ptr ms,
				  int *clause_lits, int num_lits)
{
  int i;
  vec<Lit> cl;
  for(i = 0; i < num_lits; i++) {
    const int lit = clause_lits[i];
    assert(abs(lit) > 0);
    assert(abs(lit) <= MiniSat_Nof_Variables((Solver*)ms));
    int var = abs(lit) - 1;
    cl.push((lit > 0) ? Lit(var) : ~Lit(var));
  }
  ((Solver *)ms)->addClause(cl);
  if(((Solver *)ms)->okay())
    return 1;
  return 0;
}

extern "C" int MiniSat_Solve(MiniSat_ptr ms)
{
  bool ret = ((Solver *)ms)->solve();
  if(ret)
    return 1;
  return 0;
}


/*
 * Here the assumption is in "dimacs form", variable indexing is 1...N
 */
extern "C" int MiniSat_Solve_Assume(MiniSat_ptr ms,
				    int nof_assumed_lits,
				    int *assumed_lits)
{
  int i;
  vec<Lit> cl;
  assert(((Solver*)0) != ((Solver*)ms)); 
  Solver& solver = *((Solver*)ms);

  solver.simplifyDB();
  if(solver.okay() == false)
    return 0;

  assert(nof_assumed_lits >= 0);
  for(i = 0; i < nof_assumed_lits; i++) {
    const int lit = assumed_lits[i];
    assert(abs(lit) > 0);
    assert(abs(lit) <= MiniSat_Nof_Variables((Solver*)ms));
    int var = abs(lit) - 1;
    cl.push((lit > 0) ? Lit(var) : ~Lit(var));
  }

  if (solver.solve(cl))
    return 1;
  return 0;
}

extern "C" int MiniSat_simplifyDB(MiniSat_ptr ms)
{
  ((Solver *)ms)->simplifyDB();
  if(((Solver *)ms)->okay())
    return 1;
  return 0;
}

/*
 * Here variables are numbered 1...N
 */
extern "C" int MiniSat_Get_Value(MiniSat_ptr ms, int var_num)
{
  assert(var_num > 0);
  if(var_num > MiniSat_Nof_Variables(ms))
    return -1;
  /* minisat assigns all variables. just check */
  assert(((Solver *)ms)->model[var_num-1] != l_Undef); 
  
  if(((Solver *)ms)->model[var_num-1] == l_True)
    return 1;
  return 0;
}

/* new for itp support */

extern "C" void MiniSat_EnableInterpolation(MiniSat_ptr ms)
{
  assert(!MiniSat_InterpolationIsEnabled(ms));

  Solver* sms = (Solver*)ms;

  sms->interpolator = new Interpolator();
  assert(sms->interpolator != (Interpolator*)NULL);

  sms->proof = new Proof(*(sms->interpolator));
  assert(sms->proof != (Proof*)NULL);
}

extern "C" int MiniSat_InterpolationIsEnabled(MiniSat_ptr ms)
{
  return
    (((Solver *)ms)->proof) != NULL
    &&
    (((Solver *)ms)->interpolator) != NULL
    ?
    1:
    0;
}

extern "C" void MiniSat_EnableVerboseProofLogging(MiniSat_ptr ms)
{
  assert(MiniSat_InterpolationIsEnabled(ms));
  Interpolator* itp = ((Solver *)ms)->interpolator;
  itp->enableVerboseProofLogging();
}

extern "C" void MiniSat_DisableVerboseProofLogging(MiniSat_ptr ms)
{
  assert(MiniSat_InterpolationIsEnabled(ms));
  Interpolator* itp = ((Solver *)ms)->interpolator;
  itp->enableVerboseProofLogging();
}

extern "C" void MiniSat_ComputeVariableLists(MiniSat_ptr ms)
{
  assert(MiniSat_InterpolationIsEnabled(ms));
  Interpolator* itp = ((Solver *)ms)->interpolator;
  itp->computeVariableLists();
}


extern "C" int MiniSat_IsVariableOfB(MiniSat_ptr ms, int literal)
{
  assert(MiniSat_InterpolationIsEnabled(ms));
  Interpolator* itp = ((Solver *)ms)->interpolator;
  return itp->isVariableOfB(literal);
}


extern "C" void MiniSat_PrintVariableLists(MiniSat_ptr ms)
{
  assert(MiniSat_InterpolationIsEnabled(ms));
  Interpolator* itp = ((Solver *)ms)->interpolator;
  itp->printVariableLists();
}

extern "C" void MiniSat_MarkProblemSplit(MiniSat_ptr ms, int i)
{
  assert(MiniSat_InterpolationIsEnabled(ms));
  Interpolator* itp = ((Solver *)ms)->interpolator;
  itp->markProblemSplit(i);
}

extern "C" struct ItpNode * MiniSat_GetInterpolantGraph (MiniSat_ptr ms)
{
  assert(MiniSat_InterpolationIsEnabled(ms));
  Interpolator* itp = ((Solver *)ms)->interpolator;

  struct ItpNode * lastItpNode = itp->getLastItpNode();

  return lastItpNode;
}

