
#include "isc.h"
#include "iscInt.h"
#include "iscAbProblem.h"
#include "iscAbProblem_private.h"

#include "bmc/bmcUtils.h"
#include "rbc/rbc.h"
#include "rbc/rbcInt.h"
#include "sat/solvers/SatMinisat.h"
#include "sat/solvers/SatMinisat_private.h"
#include "parser/symbols.h"

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static be_ptr compute_partial_transition_approx_rec ARGS((IscAbProblem_ptr self,
    struct ItpNode * itpGraph, int time));

static be_ptr compute_interpolant_rec ARGS((IscAbProblem_ptr self,
    struct ItpNode * itpGraph));
static bdd_ptr compute_interpolant_bdd_rec ARGS((IscAbProblem_ptr self,
    struct ItpNode * itpGraph, hash_ptr helpervars_map));

static void clear_tmp_results_in_itp_graph ARGS((struct ItpNode * itpGraph));
static void free_temp_bdd_in_itp_graph ARGS((struct ItpNode * itpGraph));

static void print_clauses(Be_Cnf_ptr cnf, FILE* out, SatMinisat_ptr ms);

static IscAbProblem_ptr create_ab_problem(Slist_ptr A, be_ptr B, int k,
    BeEnc_ptr be_enc);


/*---------------------------------------------------------------------------*/
/* Definition of external functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
IscAbProblem_ptr Isc_AbProblem_Create(be_ptr A, be_ptr B, int k,
    BeEnc_ptr be_enc)
{
  nusmv_assert((be_ptr)NULL != A);
  nusmv_assert((be_ptr)NULL != B);
  nusmv_assert(k > 0);

  Slist_ptr singleton = Slist_create();
  Slist_push(singleton, A);

  IscAbProblem_ptr result = create_ab_problem(singleton, B, k, be_enc);

  // cleanup
  Slist_destroy(singleton);

  return result;
}

IscAbProblem_ptr Isc_AbProblem_Create_List(Slist_ptr A, be_ptr B, int k,
    BeEnc_ptr be_enc)
{
  nusmv_assert((be_ptr)NULL != B);
  nusmv_assert(k > 0);

  return create_ab_problem(A, B, k, be_enc);
}

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
void Isc_AbProblem_Destroy(IscAbProblem_ptr self)
{
  ISC_ABPROBLEM_CHECK_INSTANCE(self);

  Siter iter;
  SLIST_FOREACH(self->cnfA, iter) {
	  Be_Cnf_ptr tmp = (Be_Cnf_ptr)Siter_element(iter);
	  Be_Cnf_Delete(tmp);
  }
  Slist_destroy(self->cnfA);

  Be_Cnf_Delete(self->cnfB);
  SatSolver_destroy(self->solver);

  FREE(self);
}

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
NodeList_ptr Isc_AbProblem_GetSharedCnfVar(IscAbProblem_ptr self)
{
  ISC_ABPROBLEM_CHECK_INSTANCE(self);

  NodeList_ptr res = NodeList_create();

  Siter iter;
  Siter inner_iter;


  //Slist_ptr varsA = Slist_create();
  Set_t varsA = Set_MakeEmpty();

  SLIST_FOREACH(self->cnfA, iter) {
	  Be_Cnf_ptr tmp = (Be_Cnf_ptr)Siter_element(iter);
	  Slist_ptr tmpVars = Be_Cnf_GetVarsList(tmp);

	  //Slist_append(varsA, tmpVars);
	  SLIST_FOREACH(tmpVars, inner_iter) {
		  nusmv_ptrint cnfLit = (nusmv_ptrint)Siter_element(inner_iter);
		  varsA = Set_AddMember(varsA, (Set_Element_t) cnfLit);
	  }
  }

  Slist_ptr varsB = Be_Cnf_GetVarsList(self->cnfB);
  
  SLIST_FOREACH(varsB, iter) {
    nusmv_ptrint cnfLit = (nusmv_ptrint)Siter_element(iter);
    if (Set_IsMember(varsA, (Set_Element_t)cnfLit)) {
      NodeList_append(res, NODE_FROM_INT(cnfLit));
    }
  }

  Set_ReleaseSet(varsA);

  return res;
}

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
be_ptr Isc_AbProblem_GetBeProblem(IscAbProblem_ptr self)
{
  ISC_ABPROBLEM_CHECK_INSTANCE(self);
  return self->beProb;
}

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
int Isc_AbProblem_GetNumSharedVariables(IscAbProblem_ptr self)
{
  ISC_ABPROBLEM_CHECK_INSTANCE(self);

  NodeList_ptr shared_cnf_vars = Isc_AbProblem_GetSharedCnfVar(self);
  int num_shared_cnf_vars = NodeList_get_length(shared_cnf_vars);
  NodeList_destroy(shared_cnf_vars);

  return num_shared_cnf_vars;
}

/**Function********************************************************************

   Synopsis           [Prints the DIMACS problem plus additional information.]

   Description        [Verbose version also prints lists of clauses, variables,
                       and variables meanings.]

   SideEffects        []

   SeeAlso            [Bmc_Dump_DimacsProblem]

******************************************************************************/
void Isc_AbProblem_DumpProblem(IscAbProblem_ptr self, FILE* out)
{

  int max_cnf_index;
  int num_clauses;

  ISC_ABPROBLEM_CHECK_INSTANCE(self);
  nusmv_assert(NULL != out);

  fprintf(out, "c ISC AB Problem Dump\n");

  /* dump cnf variable descriptions */

  Isc_AbProblem_DumpCnfDescription(self, out);

  /* Prints the problem header. */

//  max_cnf_index = max(Be_Cnf_GetMaxVarIndex(self->cnfA),
//                      Be_Cnf_GetMaxVarIndex(self->cnfB));

//  num_clauses = Be_Cnf_GetClausesNumber(self->cnfA) +
//                Be_Cnf_GetClausesNumber(self->cnfB) +
//                2; /* the two formula literals */

//  fprintf(out, "p cnf %d %d\n", max_cnf_index, num_clauses);

  /* Print clauses and formula literals. */

  Siter siter;
  int cnt = 1;
  SLIST_FOREACH(self->cnfA, siter) {
	  Be_Cnf_ptr tmp = (Be_Cnf_ptr)Siter_element(siter);
	  fprintf(out, "c Clauses of A_%d\n", cnt);
	  fprintf(out, "%d 0\n", Be_Cnf_GetFormulaLiteral(tmp));
	  print_clauses(tmp, out, SAT_MINISAT(NULL));
	  cnt++;
  }

  fprintf(out, "c Clauses of B\n");
  fprintf(out, "%d 0\n", Be_Cnf_GetFormulaLiteral(self->cnfB));
  print_clauses(self->cnfB, out, SAT_MINISAT(NULL));
}

/**Function********************************************************************

   Synopsis           [Prints the meaning of some CNF variables.]

   Description        []

   SideEffects        []

   SeeAlso            [Bmc_Dump_DimacsProblem]

******************************************************************************/
void Isc_AbProblem_DumpCnfDescription(IscAbProblem_ptr self, FILE* out)
{
  int time;
  int iter;
  int be_index;
  int cnf_index;

  ISC_ABPROBLEM_CHECK_INSTANCE(self);
  nusmv_assert(NULL != out);

  for (time = 0; time <= self->k; ++time) {
    fprintf(out, "c \nc @@@@@ Time %d\n", time);

    iter = BeEnc_get_first_untimed_var_index(self->be_enc,
             BE_VAR_TYPE_CURR | BE_VAR_TYPE_INPUT);
    while (BeEnc_is_var_index_valid(self->be_enc, iter)) {
      /* to avoid dumping of input variables at time k */
      if ((!BeEnc_is_index_input_var(self->be_enc, iter)) || (time < self->k)) {

        be_index = BeEnc_var_to_index(self->be_enc,
                                BeEnc_index_to_timed(self->be_enc, iter, time));

        cnf_index = Be_BeIndex2CnfLiteral(self->be_mgr, be_index);

        if (cnf_index != 0) {

          int beLit = Be_CnfLiteral2BeLiteral(self->be_mgr, cnf_index);
          int beIdx = Be_BeLiteral2BeIndex(self->be_mgr, beLit);

          nusmv_assert(beIdx == be_index);

          int ms_index =
            sat_minisat_cnfLiteral2minisatLiteral(SAT_MINISAT(self->solver), cnf_index);

          fprintf(out, "c CNF VAR %d (BE VAR %d, MS VAR %d) => Time %d, Model Variable ",
              cnf_index, be_index, ms_index, time);
          print_node(out, BeEnc_index_to_name(self->be_enc, iter));
          fprintf(out, "\n");
        }
      }

      iter = BeEnc_get_next_var_index(self->be_enc, iter,
              BE_VAR_TYPE_CURR | BE_VAR_TYPE_INPUT);
    }
  } /* time cycle */
  fprintf(out, "c \n");

  /* print variables of A */

  Siter siter, oiter;
  int cnt = 1;
  SLIST_FOREACH(self->cnfA, oiter) {
	  Be_Cnf_ptr tmp = (Be_Cnf_ptr)Siter_element(oiter);
	  Slist_ptr varsAi = Be_Cnf_GetVarsList(tmp);
	  fprintf(out, "c variables of A_%d:", cnt);

	  SLIST_FOREACH(varsAi, siter) {
	    nusmv_ptrint cnfLit = (nusmv_ptrint)Siter_element(siter);
	    fprintf(out, " %d", cnfLit);
	  }
	  fprintf(out, "\n");
	  cnt++;
  }


  //Slist_ptr varsA = Be_Cnf_GetVarsList(self->cnfA);
  //Siter siter;

  //fprintf(out, "c variables of A:");
  //SLIST_FOREACH(varsA, siter) {
    //nusmv_ptrint cnfLit = (nusmv_ptrint)Siter_element(siter);
    //fprintf(out, " %d", cnfLit);
  //}
  //fprintf(out, "\n");



  /* print variables of B */

  Slist_ptr varsB = Be_Cnf_GetVarsList(self->cnfB);

  fprintf(out, "c variables of B:");
  SLIST_FOREACH(varsB, siter) {
    nusmv_ptrint cnfLit = (nusmv_ptrint)Siter_element(siter);
    fprintf(out, " %d", cnfLit);
  }
  fprintf(out, "\n");

  /* print shared variables */
  fprintf(out, "c shared variables:");

  NodeList_ptr sharedVars = Isc_AbProblem_GetSharedCnfVar(self);

  ListIter_ptr lsiter;
  NODE_LIST_FOREACH(sharedVars, lsiter) {
    node_ptr el = NodeList_get_elem_at(sharedVars, lsiter);
    fprintf(out, " %d", NODE_TO_INT(el));
  }
  fprintf(out, "\nc \n");

  NodeList_destroy(sharedVars);
}

/**Function********************************************************************

   Synopsis           []

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
SatSolverResult Isc_AbProblem_Solve(IscAbProblem_ptr self, Trace_ptr* trace,
                                    NodeList_ptr trace_symbols)
{
  ISC_ABPROBLEM_CHECK_INSTANCE(self);
  nusmv_assert(SAT_SOLVER_UNAVAILABLE == self->result);

  Siter iter;
  nusmv_assert(0 < Slist_get_size(self->cnfA));


  /*
   * Note that the reverse traversal of the list is necessary to ensure
   * that variableOfAi is mapped to the lowest T^i!!!!
   */
  //int i = (self->k == Slist_get_size(self->cnfA)) ? 0 : -1;
  int i = self->k  - 1;
  Slist_reverse(self->cnfA);

  SLIST_FOREACH(self->cnfA, iter) {
	  Be_Cnf_ptr tmp = (Be_Cnf_ptr)Siter_element(iter);
	  nusmv_assert(tmp != NULL);
	  SatMinisat_MarkProblemSplit(SAT_MINISAT(self->solver), i);
	  SatSolver_add(self->solver, tmp, SatSolver_get_permanent_group(self->solver));
	  SatSolver_set_polarity(self->solver, tmp, 1,
	                          SatSolver_get_permanent_group(self->solver));
	  //i = i + 1;
	  i = i - 1;
  }

  //printf("adding cnfB...\n");
   /* mark split (to distinguish variables from A and B) */
  i = -2; // -2 is assigned to clauses in B
  SatMinisat_MarkProblemSplit(SAT_MINISAT(self->solver), i);

  SatSolver_add(self->solver, self->cnfB, SatSolver_get_permanent_group(self->solver));
  SatSolver_set_polarity(self->solver, self->cnfB, 1,
                          SatSolver_get_permanent_group(self->solver));

  //printf("computing variable lists...\n");
  SatMinisat_ComputeVariableLists(SAT_MINISAT(self->solver));

  /* abproblem is only used for interpolation, which doesn't make sense if there
   * are no shared variables. */
  //nusmv_assert(Isc_AbProblem_GetNumSharedVariables(self) > 0);

  /* SAT invokation */
  //printf("Invoking SAT-solver...\n");
  self->result = SatSolver_solve_all_groups(self->solver);

  /* Processes the result: */
  switch (self->result) {

  case SAT_SOLVER_UNSATISFIABLE_PROBLEM:
    {
      break;
    }
  case SAT_SOLVER_SATISFIABLE_PROBLEM:
    {
      if (opt_counter_examples(OptsHandler_get_instance()) &&
          (Trace_ptr*)NULL != trace) {
        *trace = Bmc_Utils_generate_cntexample(self->be_enc, self->solver,
                                              Isc_AbProblem_GetBeProblem(self),
                                              self->k, "ISC Counterexample",
                                              trace_symbols);
      }
      break;
    }
    break;

  case SAT_SOLVER_INTERNAL_ERROR:
    {
      internal_error("Sorry, solver answered with a fatal Internal "
                      "Failure during problem solving.\n");
      break;
    }
  case SAT_SOLVER_TIMEOUT:
  case SAT_SOLVER_MEMOUT:
    {
      internal_error("Sorry, solver ran out of resources and aborted "
                      "the execution.\n");
      break;
    }
  default:
    {
      internal_error("Isc: Unexpected value in sat result.\n");
      break;
    }
  } /* switch */

  return self->result;
}

/**Function********************************************************************

   Synopsis           [Computes the be representation of the interpolant.]

   Description        [All variables in the proof graph are supposed to be timed
                       at 1, and will all be timed at 0 in the resulting
                       interpolant.]

   SideEffects        []

   SeeAlso            []

******************************************************************************/
be_ptr Isc_AbProblem_GetInterpolant(IscAbProblem_ptr self)
{
  ISC_ABPROBLEM_CHECK_INSTANCE(self);
  nusmv_assert(SAT_SOLVER_UNSATISFIABLE_PROBLEM == self->result);

  if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
    printf("computing BE interpolant ...\n");
  }

  struct ItpNode * itpGraph =
    SatMinisat_GetInterpolantGraph(SAT_MINISAT(self->solver));

  be_ptr itp = compute_interpolant_rec(self, itpGraph);
  nusmv_assert(!Be_IsFalse(self->be_mgr, itp));

  #if ISC_DEBUG
  {
    /* use this hack to check if res is a valid be object */
    Be_Var2Index(self->be_mgr, itp);
  }
  #endif

  if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
    printf("computing BE interpolant ... done\n");
    if (Be_IsTrue(self->be_mgr, itp)) {
      printf("interpolant is TRUE\n");
    }
  }

  return itp;
}


/**Function********************************************************************

   Synopsis           [Computes an overapproximation of the approximate
                       transition relation.]

   Description        [Traverses the unsat proof graph and quantifies away
                       locally the unwanted (shared) variables, based on the
                       parameter 'time'.]

   SideEffects        []

   SeeAlso            []

******************************************************************************/
be_ptr Isc_AbProblem_GetPartialTransitionApproximation(IscAbProblem_ptr self,
    int time)
{
  ISC_ABPROBLEM_CHECK_INSTANCE(self);
  nusmv_assert(SAT_SOLVER_UNSATISFIABLE_PROBLEM == self->result);

  if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
    printf("computing partial transition approximation for interval (%d,%d) "
        "...\n", time, time+1);
  }

  struct ItpNode * itpGraph =
    SatMinisat_GetInterpolantGraph(SAT_MINISAT(self->solver));

  be_ptr res = compute_partial_transition_approx_rec(self, itpGraph, time);

  #if ISC_DEBUG
  {
    /* use this hack to check if res is a valid be object */
    Be_Var2Index(self->be_mgr, res);
  }
  #endif

  if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
    printf("computing partial transition approximation for interval (%d,%d) "
        "... done\n", time, time+1);
    if (Be_IsTrue(self->be_mgr, res)) {
      printf("partial transition approximation is TRUE\n");
    }
  }

  clear_tmp_results_in_itp_graph(itpGraph);

  return res;
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

   Synopsis           [Clears unsat proof graph from temporary interpolation
                       results.]

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
static void clear_tmp_results_in_itp_graph(struct ItpNode * itpGraph)
{
  if ((void*)NULL != itpGraph->interpolant) {
    itpGraph->interpolant = (void*)NULL;
    if (itpGraph->nType != ROOT_A && itpGraph->nType != ROOT_B) {
      clear_tmp_results_in_itp_graph(itpGraph->parentPositive);
      clear_tmp_results_in_itp_graph(itpGraph->parentNegative);
    }
  }
}

/**Function********************************************************************

   Synopsis           [Computes an overapproximation of the approximate
                       transition relation.]

   Description        []

   SideEffects        []

   SeeAlso            [Isc_AbProblem_GetPartialTransitionApproximation]

******************************************************************************/
static be_ptr compute_partial_transition_approx_rec (IscAbProblem_ptr self,
    struct ItpNode * itpGraph, int time)
{
  ISC_ABPROBLEM_CHECK_INSTANCE(self);
  nusmv_assert((struct ItpNode *)NULL != itpGraph);

  be_ptr aptPos;
  be_ptr aptNeg;
  be_ptr result;

  if ((void*)NULL != itpGraph->interpolant) {
    return (be_ptr)itpGraph->interpolant;
  }

  switch (itpGraph->nType) {

  case ROOT_A:
    {
//       	if (itpGraph->k == time) {
////    		//printf("-");
//    		result = Be_Falsity(self->be_mgr);
//    		for (int i = 0; i < itpGraph->clause_size; i++) {
//    			int literal = itpGraph->clause[i];
//    			int isInB = SatMinisat_IsVariableOfB(SAT_MINISAT(self->solver), literal);
//    			if (isInB) {
//                    int cnfLit = sat_minisat_minisatLiteral2cnfLiteral(
//                            SAT_MINISAT(self->solver), (literal >> 1)+1);
//                    nusmv_assert(cnfLit > 0);
//
//                    int beLit = Be_CnfLiteral2BeLiteral(self->be_mgr, cnfLit);
//                    nusmv_assert(beLit > 0);
//
//                    int beIdx = Be_BeLiteral2BeIndex(self->be_mgr, beLit);
//                    nusmv_assert(beIdx > 0);
//
//                    int beIdx_time = BeEnc_index_to_time(self->be_enc, beIdx);
//                    nusmv_assert(beIdx_time >= 0);
//
//                    int beIdx_ut = BeEnc_index_to_untimed_index(self->be_enc, beIdx);
//
//                    be_ptr beVar = BeEnc_index_to_var(self->be_enc, beIdx_ut);
//
//                    //printf("time %d\t beIdx %d\t beIdx_ut %d\t beIdx_time %d\n", time, beIdx, beIdx_ut, beIdx_time);
//
//                    /* cast to next-state variable? */
//                    if (beIdx_time == itpGraph->k+1) {
//                    	nusmv_assert(!BeEnc_is_index_input_var(self->be_enc, beIdx_ut));
//                    	nusmv_assert(BeEnc_is_index_state_var(self->be_enc, beIdx_ut));
//                    	beVar = BeEnc_var_curr_to_next(self->be_enc, beVar);
//                        nusmv_assert(BeEnc_is_index_untimed_next(self->be_enc,
//                            BeEnc_var_to_index(self->be_enc, beVar)));
//                    } else {
//                    	nusmv_assert(BeEnc_var_to_index(self->be_enc, beVar) == beIdx_ut);
//                    	nusmv_assert(
//                    	              BeEnc_is_index_untimed_curr(self->be_enc,
//                    	                BeEnc_var_to_index(self->be_enc, beVar)) ||
//                    	              BeEnc_is_index_untimed_input(self->be_enc,
//                    	                BeEnc_var_to_index(self->be_enc, beVar)));
//                    }
//
//                    // sign is negative
//                    if ((literal & 1)) {
//                    	//printf("-=-");
//                    	beVar = Be_Not(self->be_mgr, beVar);
//                    }
//
//                    result = Be_Or(self->be_mgr, result, beVar);
//    			}
//    		}
//    	} else {
//    		//printf("+");
//    		result = Be_Truth(self->be_mgr);
//    	}


    	if (itpGraph->k == time) {
    		result = Be_Falsity(self->be_mgr);
    	} else {
    		result = Be_Truth(self->be_mgr);
    	}
      break;
    }

  case ROOT_B:
    {
      //printf("+");

      result = Be_Truth(self->be_mgr);
      break;
    }

  case PIVOT_A:
    {
      aptPos = compute_partial_transition_approx_rec(self,
          itpGraph->parentPositive, time);
      aptNeg = compute_partial_transition_approx_rec(self,
          itpGraph->parentNegative, time);

      if ((itpGraph->k == time) || (itpGraph->k == time+1)) {
    	  result = Be_Or(self->be_mgr, aptPos, aptNeg);
      } else {
    	  result = Be_And(self->be_mgr, aptPos, aptNeg);
      }
      break;


//      int pivotVar_t = (int)(itpGraph->pivotVariable);
//      nusmv_assert(pivotVar_t > 0);
//
//      int cnfLit_t = sat_minisat_minisatLiteral2cnfLiteral(
//    		  SAT_MINISAT(self->solver), pivotVar_t);
//      nusmv_assert(cnfLit_t > 0);
//
//      int testPivot = sat_minisat_cnfLiteral2minisatLiteral(
//    		  SAT_MINISAT(self->solver), cnfLit_t);
//      nusmv_assert(testPivot == pivotVar_t);
//
//      int beLit_t = Be_CnfLiteral2BeLiteral(self->be_mgr, cnfLit_t);
//
//      // hack to determine if it's a helper variable or not
//
//      if (beLit_t == 0) {
//    	  // it's a helper var
//    	  //result = Be_And(self->be_mgr, aptPos, aptNeg);
//          if (itpGraph->k == time || itpGraph->k == time+1) {
//        	  result = Be_Or(self->be_mgr, aptPos, aptNeg);
//          } else {
//        	  result = Be_And(self->be_mgr, aptPos, aptNeg);
//          }
//      } else {
//    	  // it a non-helper var
//    	  int beIdx_t = Be_BeLiteral2BeIndex(self->be_mgr, beLit_t);
//    	  nusmv_assert(beIdx_t > 0);
//
//    	  int beIdx_time = BeEnc_index_to_time(self->be_enc, beIdx_t);
//    	  nusmv_assert(beIdx_time >= 0);
//
//          if (beIdx_time == time || beIdx_time == time+1) {
//    	  //if (beIdx_time == time) {
//        	  result = Be_Or(self->be_mgr, aptPos, aptNeg);
//          } else {
//        	  result = Be_And(self->be_mgr, aptPos, aptNeg);
//          }
//      }
//      break;
    }

  case PIVOT_AB:
      {
        be_ptr aptPos = compute_partial_transition_approx_rec(self,
            itpGraph->parentPositive, time);
        be_ptr aptNeg = compute_partial_transition_approx_rec(self,
            itpGraph->parentNegative, time);

        int pivotVar_t = (int)(itpGraph->pivotVariable);
        nusmv_assert(pivotVar_t > 0);

        int cnfLit_t = sat_minisat_minisatLiteral2cnfLiteral(
            SAT_MINISAT(self->solver), pivotVar_t);
        nusmv_assert(cnfLit_t > 0);

        int testPivot = sat_minisat_cnfLiteral2minisatLiteral(
            SAT_MINISAT(self->solver), cnfLit_t);
        nusmv_assert(testPivot == pivotVar_t);

        int beLit_t = Be_CnfLiteral2BeLiteral(self->be_mgr, cnfLit_t);
        nusmv_assert(beLit_t > 0);

        int beIdx_t = Be_BeLiteral2BeIndex(self->be_mgr, beLit_t);
        nusmv_assert(beIdx_t > 0);

        int beIdx_time = BeEnc_index_to_time(self->be_enc, beIdx_t);
        nusmv_assert(beIdx_time >= 0);

        int beIdx_ut = BeEnc_index_to_untimed_index(self->be_enc, beIdx_t);

        //printf("#");

        if (beIdx_time != time && beIdx_time != time+1) {
          /* eliminate variable */
          result = Be_And(self->be_mgr, aptPos, aptNeg);
        }
        else if (beIdx_time == time+1 && BeEnc_is_index_input_var(self->be_enc, beIdx_ut)) {
          /* eliminate variable */
          result = Be_And(self->be_mgr, aptPos, aptNeg);
        }
        else {
          nusmv_assert(BeEnc_is_index_untimed_curr(self->be_enc, beIdx_ut) ||
                       BeEnc_is_index_untimed_input(self->be_enc, beIdx_ut));

          be_ptr beVar = BeEnc_index_to_var(self->be_enc, beIdx_ut);

          /* cast to next-state variable? */
          if (beIdx_time == time+1) {
            nusmv_assert(BeEnc_is_index_state_var(self->be_enc, beIdx_ut));
            beVar = BeEnc_var_curr_to_next(self->be_enc, beVar);
            nusmv_assert(BeEnc_is_index_untimed_next(self->be_enc,
                BeEnc_var_to_index(self->be_enc, beVar)));
          }
          else {
            nusmv_assert(BeEnc_var_to_index(self->be_enc, beVar) == beIdx_ut);
            nusmv_assert(
                BeEnc_is_index_untimed_curr(self->be_enc,
                  BeEnc_var_to_index(self->be_enc, beVar)) ||
                BeEnc_is_index_untimed_input(self->be_enc,
                  BeEnc_var_to_index(self->be_enc, beVar)));
          }
          /* build expression */
          be_ptr pivotPositive = beVar;
          be_ptr pivotNegative = Be_Not(self->be_mgr, beVar);
          result = Be_Or(self->be_mgr,
                        Be_And(self->be_mgr, pivotNegative, aptPos),
                        Be_And(self->be_mgr, pivotPositive, aptNeg));
        }

        break;
      }


  case PIVOT_B:
    {
      aptPos = compute_partial_transition_approx_rec(self,
          itpGraph->parentPositive, time);
      aptNeg = compute_partial_transition_approx_rec(self,
          itpGraph->parentNegative, time);
      result = Be_And(self->be_mgr, aptPos, aptNeg);
      break;
    }


  default:
    error_unreachable_code();

  }

  itpGraph->interpolant = (void*)result;
  return result;
}

/**Function********************************************************************

   Synopsis           [Frees temporary bdds of interpolant.]

   Description        [Should be called only within
                       Isc_AbProblem_GetInterpolantBdd.]

   SideEffects        []

   SeeAlso            []

******************************************************************************/
static void free_temp_bdd_in_itp_graph(struct ItpNode * itpGraph)
{
  if ((void*)NULL != itpGraph->interpolant) {

    bdd_free(dd_manager, (bdd_ptr)(itpGraph->interpolant));
    itpGraph->interpolant = (void*)NULL;

    if (itpGraph->nType != ROOT_A && itpGraph->nType != ROOT_B) {
      free_temp_bdd_in_itp_graph(itpGraph->parentPositive);
      free_temp_bdd_in_itp_graph(itpGraph->parentNegative);
    }
  }
}



/**Function********************************************************************

   Synopsis           [Recursive functions to compute interpolant.]

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
static be_ptr compute_interpolant_rec(IscAbProblem_ptr self,
    struct ItpNode * itpGraph)
{
  ISC_ABPROBLEM_CHECK_INSTANCE(self);
  nusmv_assert((struct ItpNode *)NULL != itpGraph);

  be_ptr itpPos;
  be_ptr itpNeg;
  be_ptr result;

  if ((void*)NULL != itpGraph->interpolant) {
    return (be_ptr)itpGraph->interpolant;
  }

  switch (itpGraph->nType) {

  case ROOT_A:
    {
        result = Be_Falsity(self->be_mgr);
        for (int i = 0; i < itpGraph->clause_size; i++) {
            int literal = itpGraph->clause[i];
            int isInB = SatMinisat_IsVariableOfB(SAT_MINISAT(self->solver), literal);
            if (isInB) {
                int cnfLit = sat_minisat_minisatLiteral2cnfLiteral(
                        SAT_MINISAT(self->solver), (literal >> 1)+1);
                nusmv_assert(cnfLit > 0);

                int beLit = Be_CnfLiteral2BeLiteral(self->be_mgr, cnfLit);
                nusmv_assert(beLit > 0);

                int beIdx = Be_BeLiteral2BeIndex(self->be_mgr, beLit);
                nusmv_assert(beIdx > 0);

                int beIdx_time = BeEnc_index_to_time(self->be_enc, beIdx);
                nusmv_assert(beIdx_time >= 0);

                int beIdx_ut = BeEnc_index_to_untimed_index(self->be_enc, beIdx);

                be_ptr beVar = BeEnc_index_to_timed(self->be_enc, beIdx_ut, 0);
                //be_ptr beVar = BeEnc_index_to_var(self->be_enc, beIdx_ut);

                 // sign is negative
                if ((literal & 1)) {
                    //printf("-=-");
                    beVar = Be_Not(self->be_mgr, beVar);
                }

                result = Be_Or(self->be_mgr, result, beVar);
            }
        }

      break;
    }

  case ROOT_B:
    {
      result = Be_Truth(self->be_mgr);
      break;
    }

  case PIVOT_A:
    {
      itpPos = compute_interpolant_rec(self, itpGraph->parentPositive);
      itpNeg = compute_interpolant_rec(self, itpGraph->parentNegative);
      result = Be_Or(self->be_mgr, itpPos, itpNeg);
      break;
    }

  case PIVOT_B:
  case PIVOT_AB:
    {
      itpPos = compute_interpolant_rec(self, itpGraph->parentPositive);
      itpNeg = compute_interpolant_rec(self, itpGraph->parentNegative);
      result = Be_And(self->be_mgr, itpPos, itpNeg);
      break;
    }

    /*
  case PIVOT_AB:
    {

      int pivotVar_t1 = (int)(itpGraph->pivotVariable);
      nusmv_assert(pivotVar_t1 > 0);

      int cnfLit_t1 = sat_minisat_minisatLiteral2cnfLiteral(
          SAT_MINISAT(self->solver), pivotVar_t1);
      nusmv_assert(cnfLit_t1 > 0);

      int beLit_t1 = Be_CnfLiteral2BeLiteral(self->be_mgr, cnfLit_t1);
      nusmv_assert(beLit_t1 != 0);
      nusmv_assert(beLit_t1 > 0);

      int beIdx_t1 = Be_BeLiteral2BeIndex(self->be_mgr, beLit_t1);
      nusmv_assert(BeEnc_index_to_time(self->be_enc, beIdx_t1) == 1);
      nusmv_assert(beIdx_t1 > 0);

      int beIdx_ut = BeEnc_index_to_untimed_index(self->be_enc, beIdx_t1);

      be_ptr beVar_t0 = BeEnc_index_to_timed(self->be_enc, beIdx_ut, 0);

      be_ptr pivotPositive = beVar_t0;
      be_ptr pivotNegative = Be_Not(self->be_mgr, beVar_t0);

      be_ptr itpPos = compute_interpolant_rec(self, itpGraph->parentPositive);
      be_ptr itpNeg = compute_interpolant_rec(self, itpGraph->parentNegative);

      result = Be_Or(self->be_mgr,
                     Be_And(self->be_mgr, pivotNegative, itpPos),
                     Be_And(self->be_mgr, pivotPositive, itpNeg));
      break;
    }*/

  default:
    error_unreachable_code();

  }

  itpGraph->interpolant = (void*)result;
  return result;
}

/**Function********************************************************************

   Synopsis           [Printing functions]

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
static void print_clauses(Be_Cnf_ptr cnf, FILE* out, SatMinisat_ptr ms)
{
  Slist_ptr clauses = Be_Cnf_GetClausesList(cnf);
  nusmv_assert(Slist_get_size(clauses) > 0);

  Siter iter_clauses;
  int* cl = (int*)NULL;
  nusmv_ptrint lit = 0;
  nusmv_ptrint msLit = 0;

  SLIST_FOREACH(clauses, iter_clauses) {
    cl = (int*) Siter_element(iter_clauses);

    int i = 0;
    while (cl[i] != 0) {
      lit = (nusmv_ptrint)cl[i];
      if (SAT_MINISAT(NULL) != ms) {
        msLit = sat_minisat_cnfLiteral2minisatLiteral(ms, lit);
        fprintf(out, "%" PRIdPTR " ", lit);
        fprintf(out, "(%" PRIdPTR ") ", msLit);
      }
      else {
        fprintf(out, "%" PRIdPTR " ", lit);
      }
      i++;
    }
    fprintf(out, "0\n");
  }
}


/**Function********************************************************************

   Synopsis           [creating a SAT formula]

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
static IscAbProblem_ptr create_ab_problem(Slist_ptr A, be_ptr B, int k,
    BeEnc_ptr be_enc)
{
  IscAbProblem_ptr self = ALLOC(IscAbProblem, 1);

  /* initialize general helpers and results */

  self->bdd_enc = Enc_get_bdd_encoding();
  self->be_enc = be_enc;
  self->be_mgr = BeEnc_get_be_manager(be_enc);

  nusmv_assert(!Be_IsTrue(self->be_mgr, B));
  nusmv_assert(!Be_IsFalse(self->be_mgr, B));

  self->result = SAT_SOLVER_UNAVAILABLE;

  self->solver = Sat_CreateNonIncSolver("MiniSat");
  if (self->solver == SAT_SOLVER(NULL)) {
    internal_error("Error: This version of NuSMV must be linked with "
        "MiniSat (version 1.14p-itp).\n");
  }

  SatMinisat_EnableInterpolation(SAT_MINISAT(self->solver));
  nusmv_assert(SatMinisat_InterpolationIsEnabled(SAT_MINISAT(self->solver)));

  if (opt_verbose_level_gt(OptsHandler_get_instance(), 20)) {
    SatMinisat_EnableVerboseProofLogging(SAT_MINISAT(self->solver));
  }

  /* initialize main attributes */

  self->k = k;
  self->beA = A;
  self->beB = B;


  /* PART B */

  self->cnfB =
	Be_ConvertToCnf(self->be_mgr, B, 1);
    //Be_ConvertToCnf(self->be_mgr, Bmc_Utils_apply_inlining(self->be_mgr, B), 1);

  nusmv_assert(0 < Be_Cnf_GetClausesNumber(self->cnfB));

  self->beProb = B;

  /* reset rbc cache to avoid helper variable sharing between cnfs */

  Be_RbcManager_Reset(self->be_mgr);


  /* PART A */

  self->cnfA = Slist_create();
  Siter iter;

  SLIST_FOREACH(A, iter) {
	  be_ptr be_Ai = (be_ptr)Siter_element(iter);
	  nusmv_assert(!Be_IsTrue(self->be_mgr, be_Ai));
	  nusmv_assert(!Be_IsFalse(self->be_mgr, be_Ai));
	  self->beProb = Be_And(self->be_mgr, self->beProb, be_Ai);

	  Be_Cnf_ptr tmp =
			  Be_ConvertToCnf(self->be_mgr, be_Ai, 1);
	  	  	  //Be_ConvertToCnf(self->be_mgr, Bmc_Utils_apply_inlining(self->be_mgr, be_Ai), 1);
	  nusmv_assert(0 < Be_Cnf_GetClausesNumber(tmp));
	  // Note: this pushes tmp to the front of the list.
	  Slist_push(self->cnfA, tmp);

	  /* reset rbc cache to avoid helper variable sharing between cnfs */
	  Be_RbcManager_Reset(self->be_mgr);
  }
  //Slist_reverse(self->cnfA);
  nusmv_assert(Slist_get_size(self->cnfA) == Slist_get_size(A));


  /* print cnf statistics */
  /*
  if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
      Be_Cnf_PrintStat(self->cnfA, nusmv_stderr, "cnfA");
      Be_Cnf_PrintStat(self->cnfB, nusmv_stderr, "cnfB");
  }*/

  if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
    Isc_AbProblem_DumpProblem(self, nusmv_stdout);
  }

  return self;
}


