
#include "isc.h"
#include "iscInt.h"
#include "iscAbProblem.h"

#include "bmc/bmcConv.h"
#include "bmc/bmcModel.h"
#include "bmc/bmcUtils.h"
#include "fsm/sexp/SexpFsm.h"
#include "mc/mc.h"
#include "parser/symbols.h"
#include "prop/propPkg.h"
#include "prop/PropDb.h"
#include "sat/SatSolver.h"
#include "set/set.h"
#include "trace/TraceManager.h"
//#include "utils/utils.h"
#include "utils/assoc.h"
#include "utils/error.h"
#include "utils/ustring.h"
#include "wff/w2w/w2w.h"

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

/* property related functions */

static be_ptr isc_get_property_unrolling ARGS((Be_Manager_ptr be_mgr,
    BeEnc_ptr be_enc, be_ptr be_prop, int from, int to, boolean xor_mode));

/* environment approximation (compositional algorithm) */

static be_ptr isc_get_approx_environment_trans ARGS((IscAbProblem_ptr abProb,
    BeEnc_ptr be_enc, Be_Manager_ptr be_mgr, hash_ptr helpervars_map,
    hash_ptr varlist_map, int k));

/* for creating the fsm split */

static boolean isc_var_belongs_to_module ARGS((node_ptr var_name,
    char* module_name));

static void isc_split_be_fsm ARGS((node_ptr formula, char* module_name,
    Isc_splitmode split_mode, BeFsm_ptr* module_be_fsm, 
    BeFsm_ptr* environment_be_fsm, BeFsm_ptr monolithic_be_fsm,
    NodeList_ptr trace_symbols));

/* sat stuff */

static boolean are_logically_equivalent ARGS((BeEnc_ptr be_enc,
    Be_Manager_ptr be_mgr, be_ptr A, be_ptr B, int k, NodeList_ptr trace_symbols));
static boolean is_valid ARGS((BeEnc_ptr be_enc, Be_Manager_ptr be_mgr,
    be_ptr prob, int k, NodeList_ptr trace_symbols));
static SatSolverResult isc_core_is_sat ARGS((BeEnc_ptr be_enc,
    Be_Manager_ptr be_mgr, be_ptr beProblem, int k, Trace_ptr* trace,
    NodeList_ptr trace_symbols));

/*---------------------------------------------------------------------------*/
/* Definition of external functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

   Synopsis           [Interpolation-based model checking of invar properties.]

   Description        [This is a wrapper function for several algorithms:
                       + noitp
                       + noitpsplit
                       + standard 
                       + compositional
                       
                       Doesn't consider fairness constraints.]

   SideEffects        []

   SeeAlso            [Bmc_GenSolveInvar]

******************************************************************************/
int Isc_CheckInvar(Prop_ptr invarprop, Isc_algorithm alg, char* module_name,
    Isc_splitmode split_mode, boolean accumulate_env_trans,
    boolean xor_prop_violation)
{
  node_ptr binvarspec = Nil; /* Its booleanization */
  BeFsm_ptr be_fsm = BE_FSM(NULL); /* The corresponding be fsm */

  BoolSexpFsm_ptr bool_fsm = BOOL_SEXP_FSM(NULL);
  Trace_ptr trace = TRACE(NULL);
  NodeList_ptr trace_symbols = NODE_LIST(NULL);
  int trace_idx = -1;
  Bmc_result result = BMC_ERROR;

  /* preconditions, have to be guaranteed by caller */
  nusmv_assert(PROP(NULL) != invarprop);
  nusmv_assert(Prop_Unchecked == Prop_get_status(invarprop));
  nusmv_assert(Prop_Invar == Prop_get_type(invarprop));

  /* retrieve be fsm */

  be_fsm = Prop_get_be_fsm(invarprop); 
  if (be_fsm == BE_FSM(NULL)) {
    PropDb_set_fsm_to_master(PropPkg_get_prop_database(), invarprop);
    be_fsm = Prop_get_be_fsm(invarprop);
    BE_FSM_CHECK_INSTANCE(be_fsm);
  }

  if (BeFsm_get_fairness_list(be_fsm) != Nil) {
    fprintf(nusmv_stderr, "Fairness requirements are not supported.\n");
    return 1;
  }

  nusmv_assert(!Prop_needs_rewriting(invarprop));

  /* retrieve trace language symbols */

  bool_fsm = Prop_get_bool_sexp_fsm(invarprop);
  BOOL_SEXP_FSM_CHECK_INSTANCE(bool_fsm);
  trace_symbols = SexpFsm_get_symbols_list(SEXP_FSM(bool_fsm));

  /* booleanized, negated and NNFed formula: */

  binvarspec = Wff2Nnf(Compile_detexpr2bexpr(Enc_get_bdd_encoding(),
                                             Prop_get_expr_core(invarprop)));

  switch (alg) {

  case ISC_ALG_NOITP:
    {
      result = Isc_check_invar_noitp_algorithm(be_fsm,
                                  binvarspec,
                                  get_bmc_pb_length(OptsHandler_get_instance()),
                                  &trace,
                                  xor_prop_violation,
                                  trace_symbols);
      break;
    }
  case ISC_ALG_NOITPSPLIT:
    {
      result = Isc_check_invar_noitpsplit_algorithm(be_fsm,
                                  module_name,
                                  split_mode,
                                  binvarspec,
                                  get_bmc_pb_length(OptsHandler_get_instance()),
                                  &trace,
                                  xor_prop_violation,
                                  trace_symbols);
      break;
    }
  case ISC_ALG_STANDARD:
    {
      result = Isc_check_invar_mcmillan_algorithm(be_fsm,
                                              binvarspec,
                                              &trace,
                                              xor_prop_violation,
                                              trace_symbols);
      break;
    }
  case ISC_ALG_COMPOSITIONAL:
    {
      result = Isc_check_invar_nguyen_algorithm(be_fsm,
                                                module_name,
                                                split_mode,
                                                binvarspec,
                                                &trace,
                                                accumulate_env_trans,
                                                xor_prop_violation,
                                                trace_symbols);
      break;
    }
  default:
    {
      error_unreachable_code();
      break;
    }
  }

  switch (result) {

  case BMC_FALSE:
    {
      fprintf(nusmv_stdout, "-- ");
      print_invar(nusmv_stdout, invarprop);
      fprintf(nusmv_stdout, "is false\n");
      Prop_set_status(invarprop, Prop_False);

      if (opt_counter_examples(OptsHandler_get_instance()) && (TRACE(NULL) != trace)) {
        /* Print the trace using default plugin */
        fprintf(nusmv_stdout,
                "-- as demonstrated by the following execution sequence\n");

        trace_idx = TraceManager_register_trace(global_trace_manager, trace);
        TraceManager_execute_plugin(global_trace_manager, TRACE_OPT(NULL),
                                    TRACE_MANAGER_DEFAULT_PLUGIN,
                                    trace_idx);

        Prop_set_trace(invarprop, trace_idx);
      }
      break;
    }
  case BMC_TRUE:
    {
      fprintf(nusmv_stdout, "-- ");
      print_invar(nusmv_stdout, invarprop);
      fprintf(nusmv_stdout, "is true\n");
      Prop_set_status(invarprop, Prop_True);
      break;
    }
  case BMC_UNKNOWN:
    {
      fprintf(nusmv_stdout, "-- cannot prove the ");
      print_invar(nusmv_stdout, invarprop);
      fprintf(nusmv_stdout, "is true or false.\n");
      break;
    }
  case BMC_ERROR:
    {
      fprintf(nusmv_stderr, "Error during sat solving.\n");
      return 1;
    }
  default:
    {
      error_unreachable_code();
    }
  }

  return 0;
}

/**Function********************************************************************

   Synopsis           [Invariant checking using BMC.]

   Description        [Standard BMC approach for invariants. Simplified version
                       of Bmc_GenSolveInvar]

   SideEffects        []

   SeeAlso            [Bmc_GenSolveInvar]

******************************************************************************/
Bmc_result Isc_check_invar_noitp_algorithm(BeFsm_ptr be_fsm,
                                           node_ptr binvarspec,
                                           int max_k,
                                           Trace_ptr* trace,
                                           boolean xor_prop_violation,
                                           NodeList_ptr trace_symbols)
{
  nusmv_assert(max_k >= 0);
  BE_FSM_CHECK_INSTANCE(be_fsm);

  BeEnc_ptr be_enc = BE_ENC(NULL);
  Be_Manager_ptr be_mgr = (Be_Manager_ptr)NULL;
  SatSolverResult sat_res = SAT_SOLVER_INTERNAL_ERROR;
  Bmc_result bmc_res = BMC_UNKNOWN;
  be_ptr be_p = (be_ptr)NULL;
  be_ptr be_not_p = (be_ptr)NULL;
  be_ptr be_init = (be_ptr)NULL;

  be_enc = BeFsm_get_be_encoding(be_fsm);
  be_mgr = BeEnc_get_be_manager(be_enc);

  int k = 0;

  be_p = Bmc_Conv_Bexp2Be(be_enc, binvarspec);
  be_not_p = Be_Not(be_mgr, be_p);

  if (Be_IsFalse(be_mgr, be_not_p)) {
    printf("[invar property is always true by definition]\n");
    return BMC_TRUE;
  }

  if (Be_IsTrue(be_mgr, be_not_p)) {
    printf("[invar property is always false by definition]\n");
    return BMC_FALSE;
  }

  be_init = Bmc_Model_GetInit0(be_fsm);

  while (k <= max_k && BMC_UNKNOWN == bmc_res) {

    IscAbProblem_ptr prob = ISC_ABPROBLEM(NULL);
    be_ptr A = (be_ptr)NULL;
    be_ptr B = (be_ptr)NULL;

    /* create A&B SAT problem */

    if (k > 0)
    {
      /* A = I_{0} & T_{0,1} */

      A = Be_And(be_mgr, be_init, Bmc_Model_GetUnrolling(be_fsm, 0, 1));

      /* B = T_{1,2} & ... & T_{k-1,k} & (!p_{1} | ... | !p_{k}) */

      /* first, compute disjunction of property violations */

      B = isc_get_property_unrolling(be_mgr, be_enc, be_p, 1, k,
          xor_prop_violation);

      /* second, compute the unrolling beyond t=1 */

      if (k > 1) {
        B = Be_And(be_mgr, B, Bmc_Model_GetUnrolling(be_fsm, 1, k));
      }

      /* solve A&B SAT problem */

      prob = Isc_AbProblem_Create(A, B, k, be_enc);
      sat_res = Isc_AbProblem_Solve(prob, trace, trace_symbols);
      Isc_AbProblem_Destroy(prob);
    }

    else /* k == 0 */
    {
      nusmv_assert(0 == k);

      /* A = I_0  */
      A = be_init;

      /* B = !p_0 */
      B = BeEnc_untimed_expr_to_timed(be_enc, be_not_p, 0);

      sat_res = isc_core_is_sat(be_enc, be_mgr, Be_And(be_mgr, A, B), 0, trace, trace_symbols);
    }

    if (SAT_SOLVER_UNSATISFIABLE_PROBLEM == sat_res) {
      fprintf(nusmv_stdout,
              "-- no counterexample found with bound %d\n", k);
    }
    else if (SAT_SOLVER_SATISFIABLE_PROBLEM == sat_res) {
      bmc_res = BMC_FALSE;
    }
    else {
      error_unreachable_code();
    }

    k = k + 1;
  } /* while k <= max_k */

  return bmc_res;
}

/**Function********************************************************************

   Synopsis           [Invariant checking using BMC.]

   Description        [Same as noitp, but splits the transition relation
                       according to a given module name.]

   SideEffects        []

   SeeAlso            [Bmc_GenSolveInvar]

******************************************************************************/
Bmc_result Isc_check_invar_noitpsplit_algorithm(BeFsm_ptr be_fsm_monolithic,
                                                char* module_name,
                                                Isc_splitmode split_mode,
                                                node_ptr binvarspec,
                                                int max_k,
                                                Trace_ptr* trace,
                                                boolean xor_prop_violation,
                                                NodeList_ptr trace_symbols)
{
  nusmv_assert(max_k >= 0);
  BE_FSM_CHECK_INSTANCE(be_fsm_monolithic);

  BeEnc_ptr be_enc = BE_ENC(NULL);
  Be_Manager_ptr be_mgr = (Be_Manager_ptr)NULL;
  SatSolverResult sat_res = SAT_SOLVER_INTERNAL_ERROR;
  Bmc_result bmc_res = BMC_UNKNOWN;

  be_ptr be_p = (be_ptr)NULL;
  be_ptr be_not_p = (be_ptr)NULL;
  be_ptr be_init = (be_ptr)NULL;
  be_ptr A = (be_ptr)NULL;
  be_ptr B = (be_ptr)NULL;

  IscAbProblem_ptr abProb = ISC_ABPROBLEM(NULL);

  be_enc = BeFsm_get_be_encoding(be_fsm_monolithic);
  be_mgr = BeEnc_get_be_manager(be_enc);

  int k = 0;

  BeFsm_ptr be_fsm_module = BE_FSM(NULL);
  BeFsm_ptr be_fsm_environment = BE_FSM(NULL);

  be_p = Bmc_Conv_Bexp2Be(be_enc, binvarspec);
  be_not_p = Be_Not(be_mgr, be_p);

  if (Be_IsFalse(be_mgr, be_not_p)) {
    printf("[invar property is always true by definition]\n");
    return BMC_TRUE;
  }

  if (Be_IsTrue(be_mgr, be_not_p)) {
    printf("[invar property is always false by definition]\n");
    return BMC_FALSE;
  }

  isc_split_be_fsm(binvarspec, module_name, split_mode, &be_fsm_module,
      &be_fsm_environment, be_fsm_monolithic, trace_symbols);

  be_init = Be_And(be_mgr,
      Bmc_Model_GetInit0(be_fsm_module),
      Bmc_Model_GetInit0(be_fsm_environment));

  while (k <= max_k && BMC_UNKNOWN == bmc_res) {

    abProb = ISC_ABPROBLEM(NULL);

    /* create A&B SAT problem */

    if (k > 0)
    {
      /* A = I_{0} & T_{0,1} */

      A = Be_And(be_mgr,
                Bmc_Model_GetUnrolling(be_fsm_module, 0, 1),
                Bmc_Model_GetUnrolling(be_fsm_environment, 0, 1));

      A = Be_And(be_mgr, be_init, A);

      /* B = T_{1,2} & ... & T_{k-1,k} & (!p_{1} | ... | !p_{k}) */

      /* first, compute disjunction of property violations */

      B = isc_get_property_unrolling(be_mgr, be_enc, be_p, 1, k,
          xor_prop_violation);

      /* second, compute the unrolling beyond t=1 */

      if (k > 1) {
        be_ptr unroll = Be_And(be_mgr,
                Bmc_Model_GetUnrolling(be_fsm_module, 1, k),
                Bmc_Model_GetUnrolling(be_fsm_environment, 1, k));

        B = Be_And(be_mgr, B, unroll);

      }

      {
        be_ptr test_A = (be_ptr)NULL;
        be_ptr test_B = (be_ptr)NULL;

        test_A = Bmc_Model_GetUnrolling(be_fsm_monolithic, 0, 1);
        test_A = Be_And(be_mgr, Bmc_Model_GetInit0(be_fsm_monolithic), test_A);

        test_B = isc_get_property_unrolling(be_mgr, be_enc, be_p, 1, k,
            xor_prop_violation);
        test_B = Be_And(be_mgr, test_B, Bmc_Model_GetUnrolling(be_fsm_monolithic, 1, k));

        be_ptr test_monolithic = Be_And(be_mgr, test_A, test_B);
        be_ptr test_split = Be_And(be_mgr, A, B);

        if (test_monolithic != test_split) {
          nusmv_assert(are_logically_equivalent(be_enc, be_mgr, test_monolithic,
                test_split, -1, trace_symbols));
        }
      }

      /* solve A&B SAT problem */

      abProb = Isc_AbProblem_Create(A, B, k, be_enc);

      if (Isc_AbProblem_GetNumSharedVariables(abProb) > 0) {
        sat_res = Isc_AbProblem_Solve(abProb, trace, trace_symbols);
      }
      else {
        sat_res = isc_core_is_sat(be_enc, be_mgr,
            Isc_AbProblem_GetBeProblem(abProb), 0, trace, trace_symbols);
      }
      
      if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
        Isc_AbProblem_DumpProblem(abProb, nusmv_stdout);
      }

      Isc_AbProblem_Destroy(abProb);
    }

    else /* k == 0 */
    {
      nusmv_assert(0 == k);

      /* A = I_0  */
      A = be_init;

      /* B = !p_0 */
      B = BeEnc_untimed_expr_to_timed(be_enc, be_not_p, 0);

      sat_res = isc_core_is_sat(be_enc, be_mgr, Be_And(be_mgr, A, B), 0, trace, trace_symbols);
    }

    if (SAT_SOLVER_UNSATISFIABLE_PROBLEM == sat_res) {
      fprintf(nusmv_stdout,
              "-- no counterexample found with bound %d\n", k);
    }
    else if (SAT_SOLVER_SATISFIABLE_PROBLEM == sat_res) {
      bmc_res = BMC_FALSE;
    }
    else {
      error_unreachable_code();
    }

    k = k + 1;
  } /* while k <= max_k */

  BeFsm_destroy(be_fsm_module);
  BeFsm_destroy(be_fsm_environment);

  return bmc_res;
}

/**Function********************************************************************

   Synopsis           [Implementation of McMillan's Interpolation Algorithm
                       for invariant properties.]

   Description        [This algorithm follows the pseudocode as described in
                       the thesis of Lars Helander.]

   SideEffects        []

   SeeAlso            [Bmc_GenSolveInvar]

******************************************************************************/
Bmc_result Isc_check_invar_mcmillan_algorithm(BeFsm_ptr be_fsm,
                                              node_ptr binvarspec,
                                              Trace_ptr* trace,
                                              boolean xor_prop_violation,
                                              NodeList_ptr trace_symbols)
{
  BE_FSM_CHECK_INSTANCE(be_fsm);
  
  BeEnc_ptr be_enc = BE_ENC(NULL);
  Be_Manager_ptr be_mgr = (Be_Manager_ptr)NULL;
  SatSolverResult sat_res = SAT_SOLVER_INTERNAL_ERROR;

  be_ptr be_p = (be_ptr)NULL;
  be_ptr be_not_p = (be_ptr)NULL;
  be_ptr be_init = (be_ptr)NULL;
  be_ptr be_prob = (be_ptr)NULL;
  be_ptr be_A = (be_ptr)NULL;
  be_ptr be_B = (be_ptr)NULL;
  be_ptr be_ITP = (be_ptr)NULL;
  be_ptr be_reach_approx = (be_ptr)NULL;
  be_ptr be_new_states = (be_ptr)NULL;

  IscAbProblem_ptr abProb = ISC_ABPROBLEM(NULL);

  be_enc = BeFsm_get_be_encoding(be_fsm);
  be_mgr = BeEnc_get_be_manager(be_enc);

  int k = 0;
  int approx_depth = 0;

  be_p = Bmc_Conv_Bexp2Be(be_enc, binvarspec);
  be_not_p = Be_Not(be_mgr, be_p);

  if (Be_IsFalse(be_mgr, be_not_p)) {
    printf("[invar property is always true by definition]\n");
    return BMC_TRUE;
  }

  if (Be_IsTrue(be_mgr, be_not_p)) {
    printf("[invar property is always false by definition]\n");
    return BMC_FALSE;
  }

  /* verify base case: does invariant hold in initial states? */

  be_init = Bmc_Model_GetInit0(be_fsm);

  k = 0;

  be_prob = Be_And(be_mgr, be_init,
                    BeEnc_untimed_expr_to_timed(be_enc, be_not_p, k));

  sat_res = isc_core_is_sat(be_enc, be_mgr, be_prob, k, trace, trace_symbols);

  if (SAT_SOLVER_SATISFIABLE_PROBLEM == sat_res) {
    return BMC_FALSE;
  }
  else if (SAT_SOLVER_UNSATISFIABLE_PROBLEM == sat_res) {
    fprintf(nusmv_stdout, "-- no counterexample found with bound %d\n", k);
    fprintf(nusmv_stdout, "-- increasing k from %d to %d\n", k, k+1);
  }
  else {
    error_unreachable_code();
  }

  /* check for longer counterexamples with k > 0 */

  k = 1;

  while (true) {

    be_reach_approx = be_init; /* reachable states (overapproximation) */
    be_new_states = be_reach_approx;

    /* part B of formula: (T_{1/2} & ... & T_{k-1/k} & (!p_1 | ... | !p_k)) */
    /* needs to be recreated only when k is increased */

    /* invariant doesn't hold at one point of the trace */

    be_B = isc_get_property_unrolling(be_mgr, be_enc, be_p, 1, k,
           xor_prop_violation);

    /* unrolling from 1 to k */
    if (k > 1) {
      be_B = Be_And(be_mgr, be_B, Bmc_Model_GetUnrolling(be_fsm, 1, k));
    }

    approx_depth = 0;

    while (true) {

      /* part A of formula (reach_0 & T_{0/1})  */

      be_A = Be_And(be_mgr, be_new_states, Bmc_Model_GetUnrolling(be_fsm, 0, 1));

      /* check if bad state is reachable in overapproximated state space */

      abProb = Isc_AbProblem_Create(be_A, be_B, k, be_enc);

      #if ISC_DEBUG
      {
        NodeList_ptr sharedCnfVars = Isc_AbProblem_GetSharedCnfVar(abProb);
        ListIter_ptr iter;
        NODE_LIST_FOREACH(sharedCnfVars, iter) {
          node_ptr curSharedVar = NodeList_get_elem_at(sharedCnfVars, iter);
          int cnfLit = NODE_TO_INT(curSharedVar);
          nusmv_assert(cnfLit > 0);
          int beLit = Be_CnfLiteral2BeLiteral(be_mgr, cnfLit);
          nusmv_assert(beLit > 0);
          int beIdx = Be_BeLiteral2BeIndex(be_mgr, beLit);
          nusmv_assert(beIdx > 0);
          int time = BeEnc_index_to_time(be_enc, beIdx);
          nusmv_assert(time == 1);
        }
      }
      #endif

      sat_res = Isc_AbProblem_Solve(abProb, be_reach_approx == be_init ? trace :
          (Trace_ptr*)NULL, trace_symbols);

      if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
        Isc_AbProblem_DumpProblem(abProb, nusmv_stdout);
      }

      if (SAT_SOLVER_SATISFIABLE_PROBLEM == sat_res) {

        Isc_AbProblem_Destroy(abProb);

        fprintf(nusmv_stdout, "-- counterexample found with bound %d%c\n",
            k + approx_depth, 0 == approx_depth ? ' ' : '\'');

        if (be_reach_approx == be_init) {
          /* It's a real counterexample. */
          nusmv_assert(0 == approx_depth);
          return BMC_FALSE;
        }

        /* Possibly a spurious counterexample, break loop and increase k. */
        break;
      }
      else if (SAT_SOLVER_UNSATISFIABLE_PROBLEM == sat_res) {

        fprintf(nusmv_stdout,
            "-- no counterexample found with bound %d%c\n", k + approx_depth,
            0 == approx_depth ? ' ' : '\'');

        be_ITP = Isc_AbProblem_GetInterpolant(abProb);

        Isc_AbProblem_Destroy(abProb);

        /* check if fixpoint has been reached (if no new states were computed) */

        be_new_states = Be_And(be_mgr, be_ITP, Be_Not(be_mgr, be_reach_approx));

        sat_res = isc_core_is_sat(be_enc, be_mgr, be_new_states, -1, (Trace_ptr*)NULL, trace_symbols);

        if (SAT_SOLVER_UNSATISFIABLE_PROBLEM == sat_res) {
          /* there are no new states: fixpoint has been reached */
          return BMC_TRUE;
        }
        else if (SAT_SOLVER_SATISFIABLE_PROBLEM == sat_res) {
          be_reach_approx = Be_Or(be_mgr, be_reach_approx, be_ITP);
        }
        else {
          /* fixpoint problem must be SAT or UNSAT */
          error_unreachable_code();
        }
      }
      else {
        /* invar problem must be SAT or UNSAT */
        error_unreachable_code();
      }

      approx_depth = approx_depth + 1;
    }

    fprintf(nusmv_stdout, "-- increasing k from %d to %d\n", k, k+1);
    k = k + 1;
  }

  error_unreachable_code();
}

/**Function********************************************************************

   Synopsis           [Implementation of Nguyen's Compositional Interpolation
                       Algorithm for invariant properties.]

   Description        []

   SideEffects        []

   SeeAlso            [Bmc_GenSolveInvar]

******************************************************************************/
Bmc_result Isc_check_invar_nguyen_algorithm(BeFsm_ptr be_fsm_monolithic,
                                            char* module_name,
                                            Isc_splitmode split_mode,
                                            node_ptr binvarspec,
                                            Trace_ptr* trace,
                                            boolean accumulate_env_trans,
                                            boolean xor_prop_violation,
                                            NodeList_ptr trace_symbols)
{
  BE_FSM_CHECK_INSTANCE(be_fsm_monolithic);

  Bmc_result res = BMC_ERROR;
  BeEnc_ptr be_enc = BE_ENC(NULL);
  Be_Manager_ptr be_mgr = (Be_Manager_ptr)NULL;
  SatSolverResult sat_res = SAT_SOLVER_INTERNAL_ERROR;


  Slist_ptr be_Ai = Slist_create();
  be_ptr be_p = (be_ptr)NULL;
  be_ptr be_not_p = (be_ptr)NULL;
  be_ptr prop_violation = (be_ptr)NULL;
  be_ptr be_init = (be_ptr)NULL;
  be_ptr be_prob = (be_ptr)NULL;
  be_ptr be_A = (be_ptr)NULL;
  be_ptr be_B = (be_ptr)NULL;
  be_ptr be_reach_approx = (be_ptr)NULL;
  be_ptr be_new_states = (be_ptr)NULL;
  be_ptr be_init_A = (be_ptr)NULL;
  be_ptr be_init_B = (be_ptr)NULL;

  IscAbProblem_ptr abProb = ISC_ABPROBLEM(NULL);

  BeFsm_ptr be_fsm_module = BE_FSM(NULL);
  BeFsm_ptr be_fsm_environment = BE_FSM(NULL);
  BeFsm_ptr be_fsm_approxenv = BE_FSM(NULL);

  hash_ptr varlist_map = (hash_ptr)NULL;
  hash_ptr helpervars_map = (hash_ptr)NULL;

  int k = 0;
  int approx_depth = 0;

  be_enc = BeFsm_get_be_encoding(be_fsm_monolithic);
  be_mgr = BeEnc_get_be_manager(be_enc);

  be_ptr approx_env_trans = Be_Truth(be_mgr);

  int INTERPOLATE_INIT = 1;
  int ENABLE_TIMERS = 1;

  /* varlist_map    -> maintains separate helper variable lists indexed by
   *                   time; is used during existential quantification in
   *                   removing helper variables not belonging to the
   *                   current time interval. */
  varlist_map = new_assoc();

  /* helpervars_map -> used to check if a helper variable for a
   *                   given beIdx has been created before and
   *                   inserted into the bdd encoding and is present
   *                   in one of the lists in varlist_map. */
  helpervars_map = new_assoc();

  be_p = Bmc_Conv_Bexp2Be(be_enc, binvarspec);
  be_not_p = Be_Not(be_mgr, be_p);

  if (Be_IsFalse(be_mgr, be_not_p)) {
    printf("[invar property is always true by definition]\n");
    res = BMC_TRUE;
    goto cleanup_and_return;
  }

  if (Be_IsTrue(be_mgr, be_not_p)) {
    printf("[invar property is always false by definition]\n");
    res = BMC_FALSE;
    goto cleanup_and_return;
  }

  /* verify base case: does invariant hold in initial states? */

  be_init = Bmc_Model_GetInit0(be_fsm_monolithic);

  k = 0;

  be_prob = Be_And(be_mgr, be_init,
                    BeEnc_untimed_expr_to_timed(be_enc, be_not_p, k));

  sat_res = isc_core_is_sat(be_enc, be_mgr, be_prob, k, trace, trace_symbols);

  if (SAT_SOLVER_SATISFIABLE_PROBLEM == sat_res) {
    res = BMC_FALSE;
    goto cleanup_and_return;
  }
  else if (SAT_SOLVER_UNSATISFIABLE_PROBLEM == sat_res) {
    fprintf(nusmv_stdout, " -- no counterexample found with bound %d\n", k);
  }
  else {
    error_unreachable_code();
  }

  /* split module and environment fsm */

  isc_split_be_fsm(binvarspec, module_name, split_mode, &be_fsm_module, 
      &be_fsm_environment, be_fsm_monolithic, trace_symbols);

  be_init_B = Bmc_Model_GetInit0(be_fsm_module);
  be_init_A = Bmc_Model_GetInit0(be_fsm_environment);


  /************************************************************************/
  /*                                                                      */
  /* outer loop (environment approximation)                               */
  /*                                                                      */
  /* Here we check for longer counterexamples, with k > 0. First, a       */
  /* reachability check is done on the exact FSM. If this is successful,  */
  /* a true counterexample has been found; else, if there is some         */
  /* interaction between the environment and the module (there are some   */
  /* shared variables between environment and module), an approximate     */
  /* environment transition relation is computed, and the algorithm       */
  /* proceeds with the fixpoint computation using McMillan's algorithm.   */
  /* If there has not been any interaction between module and             */ 
  /* environment, execution returns to the start of the outer loop,       */
  /* where k is increased.                                                */
  /*                                                                      */
  /************************************************************************/

  // ENV SPLIT INITAL CONDITION: TVY
  if (INTERPOLATE_INIT) {
	  Slist_push(be_Ai, be_init_A);
  }

  while (true) {

//    if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
//			fprintf(nusmv_stdout, "outer_loop timer start\n");
//			Utils_start_timer("outer_loop");
//		}

    fprintf(nusmv_stdout, " -- increasing k from %d to %d\n", k, k+1);
    k = k + 1;

    if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
      printf("compositional: outer loop\n");
    }

    /************************************************************************/
    /*                       base check for current k                       */
    /************************************************************************/

    prop_violation = isc_get_property_unrolling(be_mgr, be_enc,
        be_p, 1, k, xor_prop_violation);

    //be_A = Bmc_Model_GetUnrolling(be_fsm_environment, 0, k);
    //Slist_reverse(be_Ai);
    Slist_push(be_Ai, Bmc_Model_GetUnrolling(be_fsm_environment, k-1, k));
    //Slist_reverse(be_Ai);

    if (INTERPOLATE_INIT) {
    	be_B = Be_And(be_mgr, be_init_B, Bmc_Model_GetUnrolling(be_fsm_module, 0, k));
    } else {
    	be_B = Be_And(be_mgr, be_init, Bmc_Model_GetUnrolling(be_fsm_module, 0, k));
    }


    be_B = Be_And(be_mgr, be_B, prop_violation);

    if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
      printf("be_A\n");
      Be_DumpSexpr(be_mgr, be_A, nusmv_stdout);
      printf("\nbe_B\n");
      Be_DumpSexpr(be_mgr, be_B, nusmv_stdout);
    }

	//	printf("create AB problem\n");
    abProb = Isc_AbProblem_Create_List(be_Ai, be_B, k, be_enc);
	//printf("AbProblem created.\n");

    #if ISC_DEBUG
    {
      /* Is the compositional problem equivalent to the monolithic one? */

      be_ptr problem_compositional = Isc_AbProblem_GetBeProblem(abProb);

      be_ptr problem_monolithic = Be_And(be_mgr,
          Bmc_Model_GetUnrolling(be_fsm_monolithic, 0, k), prop_violation);
      problem_monolithic = Be_And(be_mgr, be_init, problem_monolithic);

      nusmv_assert(are_logically_equivalent(be_enc, be_mgr, problem_compositional,
          problem_monolithic, -1, trace_symbols));
    }
    #endif

    //if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
      printf("num shared variables = %d\n", Isc_AbProblem_GetNumSharedVariables(abProb));
    //}


    if (ENABLE_TIMERS) {
    	Utils_start_timer("BMC");
    }

    if (Isc_AbProblem_GetNumSharedVariables(abProb) > 0) {
      sat_res = Isc_AbProblem_Solve(abProb, trace, trace_symbols);
    }
    else {
      sat_res = isc_core_is_sat(be_enc, be_mgr,
          Isc_AbProblem_GetBeProblem(abProb), k, trace, trace_symbols);
    }

    if (ENABLE_TIMERS) {
    	Utils_stop_timer("BMC");
    	//printf("BMC Timer k=%d: \n", k);
    	Utils_print_timer("BMC", "");
    	Utils_reset_timer("BMC");
    }


    if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
      Isc_AbProblem_DumpProblem(abProb, nusmv_stdout);
    }

    if (opt_verbose_level_gt(OptsHandler_get_instance(), 50)) {
      char fn [MAX_STR];
      sprintf(fn, "prob_%d.dimacs", k);
      FILE* f = fopen(fn,"w+");
      Isc_AbProblem_DumpProblem(abProb, f);
      fclose(f);
    }

    if (SAT_SOLVER_SATISFIABLE_PROBLEM == sat_res) {

      Isc_AbProblem_Destroy(abProb);
      fprintf(nusmv_stdout, " -- counterexample found with bound %d\n", k);

      #if ISC_DEBUG
      {
        /* Are both the monolithic and the compositional problems satisfiable
         * when expressing them as a single cnf object? */

        be_ptr problem_compositional = Isc_AbProblem_GetBeProblem(abProb);

        be_ptr problem_monolithic = Be_And(be_mgr,
            Bmc_Model_GetUnrolling(be_fsm_monolithic, 0, k), prop_violation);
        problem_monolithic = Be_And(be_mgr, be_init, problem_monolithic);

        sat_res = isc_core_is_sat(be_enc, be_mgr, problem_compositional, -1, (Trace_ptr*)NULL, trace_symbols);
        nusmv_assert(SAT_SOLVER_SATISFIABLE_PROBLEM == sat_res);

        sat_res = isc_core_is_sat(be_enc, be_mgr, problem_monolithic, -1, (Trace_ptr*)NULL, trace_symbols);
        nusmv_assert(SAT_SOLVER_SATISFIABLE_PROBLEM == sat_res);
      }
      #endif

      res = BMC_FALSE;
      goto cleanup_and_return;
    }
    else {
      fprintf(nusmv_stdout, " -- no counterexample found with bound %d\n", k);
    }

    #if ISC_DEBUG
    {
      be_ptr beProb = Isc_AbProblem_GetBeProblem(abProb);
      SatSolverResult sat_res =
        isc_core_is_sat(be_enc, be_mgr, beProb, -1, (Trace_ptr*)NULL, trace_symbols);
      nusmv_assert(SAT_SOLVER_UNSATISFIABLE_PROBLEM == sat_res);
    }
    #endif

    if (Isc_AbProblem_GetNumSharedVariables(abProb) == 0) {
      /* No interaction between environment and module, increase k. */
      continue;
    }

    /************************************************************************/
    /*                    compute approximate environment                   */
    /************************************************************************/


    if (ENABLE_TIMERS) {
    	Utils_start_timer("INTERPOLATE");
    }

    //#if ISC_DEBUG
    be_ptr prev_approx_env_trans = approx_env_trans;
    //#endif

    if (!accumulate_env_trans) {
      approx_env_trans = Be_Truth(be_mgr);
    }

    for (int t = 0; t < k; t++) {
    	be_ptr C_env_i = Isc_AbProblem_GetPartialTransitionApproximation(
    			abProb, t);
    	approx_env_trans = Be_And(be_mgr, approx_env_trans, C_env_i);

    	// DEBUG
    	if (Be_IsTrue(be_mgr, C_env_i)) {
    		printf("C_(%d,%d) is True\n", t, t+1);
    		//Be_DumpSexpr(be_mgr, C_env_i, nusmv_stdout);
    	}

#if ISC_CONSISTENCY_CHECKS
    	{
    		be_ptr be_exact_env_trans = BeEnc_untimed_expr_to_timed(be_enc,
    				BeFsm_get_trans(be_fsm_environment), 0);
    		be_ptr be_approx_env_trans = BeEnc_untimed_expr_to_timed(be_enc,
    				approx_env_trans, 0);
    		be_ptr test_exact_approx = Be_Implies(be_mgr,
    				be_exact_env_trans, be_approx_env_trans);
    		if (!is_valid(be_enc, be_mgr, test_exact_approx, 1,
    				trace_symbols)) {
    			error_unreachable_code_msg(
    					"\n\n'exact_env -> approx_env' doesn't hold "
    					"for interval (%d,%d)\n", t, t + 1);
    		} else {
    			printf("-- generated interpolant C_(%d,%d) is valid.\n", t, t+1);
    		}
    	}
#endif
    }


    // DEBUG
    if (Be_IsTrue(be_mgr, approx_env_trans)) {
    	printf("Interpolant for %d is True\n", k);
    	//Be_DumpSexpr(be_mgr, approx_env_trans, nusmv_stdout);
    }
    //nusmv_assert(!Be_IsTrue(be_mgr, approx_env_trans));
    //nusmv_assert(!Be_IsFalse(be_mgr, approx_env_trans));



    #if ISC_DEBUG
    {
      /* Assert that the old environment approximation is an overapproximation
      * of the new one. */

      if (accumulate_env_trans) {
        be_ptr test = Be_Implies(be_mgr, approx_env_trans, prev_approx_env_trans);
        nusmv_assert(is_valid(be_enc, be_mgr, test, 1, trace_symbols));
      }

      /* check if the environment approximation using local existential
      * quantification is indeed an overapproximation of the bdd-based
      * environment approximation. This check ignores previously accumulated
      * transition constraints. */

      be_ptr approx_env_trans_global = isc_get_approx_environment_trans(abProb,
          be_enc, be_mgr, helpervars_map, varlist_map, k);

      be_ptr approx_env_trans_local = Be_Truth(be_mgr);
      int t;
      for (t = 0; t < k; t++) {
        approx_env_trans_local = Be_And(be_mgr, approx_env_trans_local,
            Isc_AbProblem_GetPartialTransitionApproximation(abProb, t));
      }

      be_ptr approx_global = BeEnc_untimed_expr_to_timed(be_enc,
          approx_env_trans_global, 0);
      be_ptr approx_local = BeEnc_untimed_expr_to_timed(be_enc,
          approx_env_trans_local, 0);

      be_ptr test_old_new = Be_Implies(be_mgr, approx_global, approx_local);
      nusmv_assert(is_valid(be_enc, be_mgr, test_old_new, 1, trace_symbols));
    }
    #endif

    Isc_AbProblem_Destroy(abProb);

    #if ISC_CONSISTENCY_CHECKS
    {
      be_ptr be_exact_env_trans = BeEnc_untimed_expr_to_timed(be_enc,
          BeFsm_get_trans(be_fsm_environment), 0);
      be_ptr be_approx_env_trans = BeEnc_untimed_expr_to_timed(be_enc,
          approx_env_trans, 0);

      if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
        printf("exact env trans: ");
        Be_DumpSexpr(be_mgr, be_exact_env_trans, nusmv_stdout);
        printf("\n");
        printf("approx env trans: ");
        Be_DumpSexpr(be_mgr, be_approx_env_trans, nusmv_stdout);
        printf("\n");
      }

      /* check if the approximate environment transition relation is
       * indeed an overapproximation of the exact one */

      be_ptr test = Be_Implies(be_mgr, be_exact_env_trans, be_approx_env_trans);
      if (!is_valid(be_enc, be_mgr, test, 1, trace_symbols)) {
        error_unreachable_code_msg("\ntest 'exactEnv -> approxEnv' failed!\n");
      }

      if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
        test = Be_Implies(be_mgr, be_approx_env_trans, be_exact_env_trans);
        if (is_valid(be_enc, be_mgr, test, -1, trace_symbols)) {
          printf("approximate env trans <-> exact env trans\n");
        }
      }
    }
    #endif

    be_fsm_approxenv = BeFsm_create(be_enc,
                                   BeFsm_get_init(be_fsm_environment),
                                   BeFsm_get_invar(be_fsm_environment),
                                   approx_env_trans,
                                   Nil);

    if (ENABLE_TIMERS) {
    	Utils_stop_timer("INTERPOLATE");
    	//printf("BMC Timer k=%d: \n", k);
    	Utils_print_timer("INTERPOLATE", "");
    	Utils_reset_timer("INTERPOLATE");
    }
    
    /************************************************************************/
    /*                   inner loop (fixpoint computation)                  */
    /************************************************************************/

//    if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
//			fprintf(nusmv_stdout, "inner loop timer start\n");
//		  Utils_start_timer("inner_loop");
//		}


    if (ENABLE_TIMERS) {
    	Utils_start_timer("UMC");
    }

    if (INTERPOLATE_INIT) {
    	be_reach_approx = be_init_B; /* reachable states (overapproximation) */
    } else {
    	be_reach_approx = be_init; /* reachable states (overapproximation) */
    }

    be_new_states = be_reach_approx;

    /* part B of formula: (T_{1/2} & ... & T_{k-1/k} & (!p_1 | ... | !p_k)) */
    /* needs to be recreated only when k is increased */

    /* invariant doesn't hold at one point of the trace */

    be_B = prop_violation;

    /* unrolling from 1 to k */
    if (k > 1) {
      //be_B = Be_And(be_mgr, be_B, Bmc_Model_GetUnrolling(be_fsm_module, 1, k));
      //be_B = Be_And(be_mgr, be_B, Bmc_Model_GetUnrolling(be_fsm_approxenv, 1, k));
        be_B = Be_And(be_mgr, be_B, Bmc_Model_GetUnrolling(be_fsm_module, 1, k));
        be_B = Be_And(be_mgr, be_B, Bmc_Model_GetUnrolling(be_fsm_approxenv, 1, k));
    }

    /* approximate search (as in mcmillan algorithm) */

    approx_depth = 0;

    while (true) {

      be_ptr be_A = (be_ptr)NULL;
      be_ptr be_ITP = (be_ptr)NULL;
      IscAbProblem_ptr abProb = ISC_ABPROBLEM(NULL);
      SatSolverResult sat_res = SAT_SOLVER_INTERNAL_ERROR;

      if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
        printf("compositional: inner loop\n");
      }

      /* part A of formula (reach_0 & T_{0/1})  */
      //printf("inner Unrolling module & with new states\n");
      be_A = Be_And(be_mgr, be_new_states, Bmc_Model_GetUnrolling(be_fsm_module, 0, 1));
      //printf("inner Unrolling env\n");
      be_A = Be_And(be_mgr, be_A, Bmc_Model_GetUnrolling(be_fsm_approxenv, 0, 1));
      //printf("unrolling done\n");

      if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
        printf("be_A\n");
        Be_DumpSexpr(be_mgr, be_A, nusmv_stdout);
        printf("\nbe_B\n");
        Be_DumpSexpr(be_mgr, be_B, nusmv_stdout);
      }

      /* check if bad state is reachable in overapproximated state space */

      //printf("creating inner AB problem\n");
      abProb = Isc_AbProblem_Create(be_A, be_B, k, be_enc);

      //printf("running solver on inner AB problem\n");
      sat_res = Isc_AbProblem_Solve(abProb, (Trace_ptr*)NULL, trace_symbols);

      if (SAT_SOLVER_SATISFIABLE_PROBLEM == sat_res) {
        Isc_AbProblem_Destroy(abProb);
        fprintf(nusmv_stdout, " -- counterexample found with bound %d'\n",
            k + approx_depth);

        if (ENABLE_TIMERS) {
        	Utils_stop_timer("UMC");
        	Utils_print_timer("UMC", "");
        	Utils_reset_timer("UMC");
        }

        break;
      }
      else if (SAT_SOLVER_UNSATISFIABLE_PROBLEM == sat_res) {

        fprintf(nusmv_stdout,
            " -- no counterexample found with bound %d'\n", k + approx_depth);

        //printf(" -- getting interpolant: ");
        be_ITP = Isc_AbProblem_GetInterpolant(abProb);
        //printf("DONE\n");

        //printf(" -- destroying AbProb: ");
        Isc_AbProblem_Destroy(abProb);
        //printf("DONE\n");

        /* check if fixpoint has been reached (if no new states were computed) */

        //printf(" -- checking if fix-point has been reached: ");
        //fflush(stdout);

        //be_new_states = be_ITP;
        be_ptr test_convergence = Be_And(be_mgr, be_ITP, Be_Not(be_mgr, be_reach_approx));
        be_new_states = test_convergence;

        //printf("created be_new_states... ");
        //fflush(stdout);

        sat_res = isc_core_is_sat(be_enc, be_mgr, test_convergence, -1, (Trace_ptr*)NULL, trace_symbols);


        if (SAT_SOLVER_UNSATISFIABLE_PROBLEM == sat_res) {
          /* there are no new states: fixpoint has been reached */
            //printf("YES\n");

            if (ENABLE_TIMERS) {
            	Utils_stop_timer("UMC");
            	Utils_print_timer("UMC", "");
            	Utils_reset_timer("UMC");
            }


        	fprintf(nusmv_stdout,
          	  " -- inner fixed point reached at depth %d'\n", k + approx_depth);

          res = BMC_TRUE;
          BeFsm_destroy(be_fsm_approxenv);
          goto cleanup_and_return;
        }
        else if (SAT_SOLVER_SATISFIABLE_PROBLEM == sat_res) {
          //printf("NO, starting performing OR...\n");
          be_reach_approx = Be_Or(be_mgr, be_reach_approx, be_ITP);
          //printf("OR performed\n");
        }
        else {
          /* fixpoint problem must be SAT or UNSAT */
          error_unreachable_code();
        }
      }
      else {
        /* invar problem must be SAT or UNSAT */
        error_unreachable_code();
      }

      approx_depth = approx_depth + 1;
    }

//    if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
//			fprintf(nusmv_stdout, "inner loop timer stop\n");
//		  Utils_stop_timer("inner_loop");
//		}

    BeFsm_destroy(be_fsm_approxenv);
  }

  error_unreachable_code();

cleanup_and_return:
//  if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
//	  Utils_stop_timer("inner_loop");
//	  Utils_stop_timer("outer_loop");
//	  Utils_print_timer("outer_loop", "");
//	  Utils_print_timer("inner_loop", "");
//	}

  if (BE_FSM(NULL) != be_fsm_module) BeFsm_destroy(be_fsm_module);
  if (BE_FSM(NULL) != be_fsm_environment) BeFsm_destroy(be_fsm_environment);

  node_ptr all_varlist_map_keys = assoc_get_keys(varlist_map, true);
  while (all_varlist_map_keys != Nil) {
    node_ptr key = car(all_varlist_map_keys);
    NodeList_ptr varlist = NODE_LIST(find_assoc(varlist_map, key));
    NodeList_destroy(varlist);
    all_varlist_map_keys = cdr(all_varlist_map_keys);
  }

  Slist_destroy(be_Ai);
  
  free_assoc(varlist_map);
  free_assoc(helpervars_map);

  return res;
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/



/**Function********************************************************************

   Synopsis           [Computes a disjunction of property violations for a
                       specific time interval.]

   Description        [Builds a formula checking for the property violation
                       starting at time 'from' up to and including time 'to'.]

   SideEffects        []

   SeeAlso            []

******************************************************************************/
static be_ptr isc_get_property_unrolling(Be_Manager_ptr be_mgr, BeEnc_ptr be_enc,
    be_ptr be_prop, int from, int to, boolean xor_mode)
{
  nusmv_assert(from <= to);

  be_ptr result = Be_Falsity(be_mgr);
  be_ptr be_not_prop = Be_Not(be_mgr, be_prop);
  be_ptr partial;
  int i;
  int j;

  if (from == to) {
    return BeEnc_untimed_expr_to_timed(be_enc, be_not_prop, from);
  }

  for (i=from; i <= to; i++) {
    partial = BeEnc_untimed_expr_to_timed(be_enc, be_not_prop, i);

    if (xor_mode) {
      for (j=from; j <= to; j++) {
        if (j == i) continue;
        partial = Be_And(be_mgr, partial, BeEnc_untimed_expr_to_timed(be_enc,
              be_prop, j));
      }
    }
    
    result = Be_Or(be_mgr, result, partial);
  }

  return result;
}

/**Function********************************************************************

   Synopsis           [Checks if a given variable belongs to a certain module.]

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
static boolean isc_var_belongs_to_module(node_ptr var_name, char* module_name)
{
	char* name_str = sprint_node(car(var_name));
  return strcmp(name_str, module_name) == 0;
}

/**Function********************************************************************

   Synopsis           [Creates an FSM for the module and an FSM for the
                       environments (all other modules).]

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
static void isc_split_be_fsm(node_ptr formula, char* module_name,
    Isc_splitmode split_mode, BeFsm_ptr* module_be_fsm,
    BeFsm_ptr* environment_be_fsm, BeFsm_ptr monolithic_be_fsm,
    NodeList_ptr trace_symbols)
{
  nusmv_assert(ISC_SPLITMODE_NONE != split_mode || strlen(module_name) > 0);

  ListIter_ptr niter;

  SexpFsm_ptr monolithic_bool_sexp_fsm =
      SEXP_FSM(PropDb_master_get_bool_sexp_fsm(PropPkg_get_prop_database()));

  BeEnc_ptr be_enc = Enc_get_be_encoding();
  BoolEnc_ptr bool_enc = Enc_get_bool_encoding();
  BddEnc_ptr bdd_enc = Enc_get_bdd_encoding();
  Be_Manager_ptr be_mgr = BeEnc_get_be_manager(be_enc);
  SymbTable_ptr symb_table = Compile_get_global_symb_table();

  SymbLayer_ptr det_layer = SymbTable_create_layer(symb_table, (char*)NULL,
                                       SYMB_LAYER_POS_BOTTOM);
  const char* det_layer_name = SymbLayer_get_name(det_layer);

  /* compute variable lists for module and environment */
  /* variables are scalar */

  NodeList_ptr vars_all = SexpFsm_get_vars_list(monolithic_bool_sexp_fsm);

  if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
    printf("all fsm variables: ");
    NodeList_print_nodes(vars_all, nusmv_stdout);
    printf("\n");
  }

  Set_t vars_module = Set_MakeEmpty();
  Set_t vars_environment = Set_MakeEmpty();
  Set_t varDeps = Set_MakeEmpty();
  Set_t formula_modules = Set_MakeEmpty();

  /* component: all variables appearing in formula */

  if (ISC_SPLITMODE_FVARIABLES == split_mode) {
    vars_module = Formula_GetDependencies(symb_table, formula, Nil);

    NODE_LIST_FOREACH(vars_all, niter) {
      node_ptr curVar = NODE_PTR(NodeList_get_elem_at(vars_all, niter));
      if (!Set_IsMember(vars_module, (Set_Element_t)curVar)) {
        vars_environment = Set_AddMember(vars_environment, (Set_Element_t)curVar);
      }
    }
  }

  /* component: all modules appearing in formula */

  else if (ISC_SPLITMODE_FMODULES == split_mode) {

    /* get formula modules */
    varDeps = Formula_GetDependencies(symb_table, formula, Nil);

    Set_Iterator_t siter;
    SET_FOREACH(varDeps, siter) {
      node_ptr curVar = Set_GetMember(varDeps, siter);
      node_ptr curPart = isc_get_root_module_name(curVar);
      if (!Set_IsMember(formula_modules, (Set_Element_t)curPart)) {
        formula_modules = Set_AddMember(formula_modules, (Set_Element_t)curPart);
      }
    }

    if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
      printf("formula modules: ");
      Set_PrintSet(nusmv_stdout, formula_modules, NULL, NULL);
      printf("\n");
    }

    /* compute variable lists */
    NODE_LIST_FOREACH(vars_all, niter) {
      node_ptr curVar = NODE_PTR(NodeList_get_elem_at(vars_all, niter));
      node_ptr curVarModule = isc_get_root_module_name(curVar);
      if (Set_IsMember(formula_modules, (Set_Element_t)curVarModule)) {
        vars_module = Set_AddMember(vars_module, (Set_Element_t)curVar);
      }
      else {
        vars_environment = Set_AddMember(vars_environment, (Set_Element_t)curVar);
      }
    }
  }

	/* component: the components appearing in a formula 
	   
		 This is different from ISC_SPLITMODE_FMODULES, as there
		 all the modules in the formula and the tree of formula's
		 is considered.
	 */

	else if (ISC_SPLITMODE_FCOMPONENT == split_mode) {

    /* get formula modules */
    varDeps = Formula_GetDependencies(symb_table, formula, Nil);

    Set_Iterator_t siter;
    SET_FOREACH(varDeps, siter) {
      node_ptr curVar = Set_GetMember(varDeps, siter);
      node_ptr curPart = car(curVar);
      if (!Set_IsMember(formula_modules, (Set_Element_t)curPart)) {
        formula_modules = Set_AddMember(formula_modules, (Set_Element_t)curPart);
      }
    }

    if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
      printf("formula modules: ");
      Set_PrintSet(nusmv_stdout, formula_modules, NULL, NULL);
      printf("\n");
    }

    /* compute variable lists */
    NODE_LIST_FOREACH(vars_all, niter) {
      node_ptr curVar = NODE_PTR(NodeList_get_elem_at(vars_all, niter));
      node_ptr curVarModule = car(curVar);
      if (Set_IsMember(formula_modules, (Set_Element_t)curVarModule)) {
        vars_module = Set_AddMember(vars_module, (Set_Element_t)curVar);
      }
      else {
        vars_environment = Set_AddMember(vars_environment, (Set_Element_t)curVar);
      }
    }
	}

  /* component: given module */

  else if (strlen(module_name) > 0) {
    NODE_LIST_FOREACH(vars_all, niter) {
      node_ptr curVar = NODE_PTR(NodeList_get_elem_at(vars_all, niter));
      if (isc_var_belongs_to_module(curVar, module_name)) {
        vars_module = Set_AddMember(vars_module, (Set_Element_t)curVar);
      }
      else {
        vars_environment = Set_AddMember(vars_environment, (Set_Element_t)curVar);
      }
    }
  }

  /* no split mode indicated? */
  else error_unreachable_code();

  nusmv_assert(Set_GiveCardinality(vars_module) > 0);

  if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
    printf("module variables: ");
    Set_PrintSet(nusmv_stdout, vars_module, NULL, NULL);
    printf("\n");
    printf("environment variables: ");
    Set_PrintSet(nusmv_stdout, vars_environment, NULL, NULL);
    printf("\n");
  }

  nusmv_assert(NodeList_get_length(vars_all) ==
               Set_GiveCardinality(vars_module) +
               Set_GiveCardinality(vars_environment));

  //if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
    printf("total variables: %d\n", NodeList_get_length(vars_all));
    printf("module variables: %d\n", Set_GiveCardinality(vars_module));
    printf("environment variables: %d\n", Set_GiveCardinality(vars_environment));
  //}

	/* create proper init expressions */
	// TODO
	/*
	Set_Iterator_t siter;
	Set_t vars_init_B = Set_Copy(vars_module);
	Set_t all_vars = SexpFsm_get_vars(monolithic_bool_sexp_fsm);
	//printf("set all vars %d\n", Set_GiveCardinality(all_vars));
  //printf("total variables: %d\n", NodeList_get_length(vars_all));

	SET_FOREACH(vars_module, siter) {
		node_ptr curVar = Set_GetMember(vars_module, siter);
		print_node(nusmv_stdout, curVar);
		printf("next expr: \n");
		Expr_ptr next_expr = SexpFsm_get_var_trans(monolithic_bool_sexp_fsm, curVar);
		//print_node(nusmv_stdout, next_expr);
    Set_t varDeps = Formula_GetDependencies(symb_table, next_expr, Nil);
	  //printf("before set vars_init_b size %d\n", Set_GiveCardinality(vars_init_B));
	  printf("before set varDeps size %d\n", Set_GiveCardinality(varDeps));
		vars_init_B = Set_Union(vars_init_B, varDeps);
	  //printf("after set vars_init_b size %d\n", Set_GiveCardinality(vars_init_B));
		//break;
	}
	
	//printf("set vars_init_b size %d\n", Set_GiveCardinality(vars_init_B));
	Set_t vars_init_A = Set_Difference(all_vars, vars_init_B);

  if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
	  Set_PrintSet(nusmv_stdout, vars_init_A, NULL, NULL);
	  Set_PrintSet(nusmv_stdout, vars_init_B, NULL, NULL);
	}

  //if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
	  printf("set all vars %d\n", Set_GiveCardinality(all_vars));
	  printf("set vars_init_a size %d\n", Set_GiveCardinality(vars_init_A));
	  printf("set vars_init_b size %d\n", Set_GiveCardinality(vars_init_B));
	//}
	*/
  /* create be fsm for module */

  BoolSexpFsm_ptr module_bool_fsm =
      FsmBuilder_create_boolean_sexp_fsm(global_fsm_builder, 
                                         mainFlatHierarchy,
                                         vars_module,
                                         bdd_enc, 
                                         det_layer);

  /* create be fsm for environment */

  BoolSexpFsm_ptr environment_bool_fsm =
      FsmBuilder_create_boolean_sexp_fsm(global_fsm_builder, 
                                         mainFlatHierarchy,
                                         vars_environment,
                                         bdd_enc, 
                                         det_layer);
	

  BaseEnc_commit_layer(BASE_ENC(bool_enc), det_layer_name);
  BaseEnc_commit_layer(BASE_ENC(be_enc), det_layer_name);
  BaseEnc_commit_layer(BASE_ENC(bdd_enc), det_layer_name);

  *module_be_fsm = BeFsm_create_from_sexp_fsm(be_enc, module_bool_fsm);
  *environment_be_fsm = BeFsm_create_from_sexp_fsm(be_enc, environment_bool_fsm);

  /* cleanup */

  Set_ReleaseSet(vars_module);
  Set_ReleaseSet(vars_environment);
  Set_ReleaseSet(formula_modules);
  Set_ReleaseSet(varDeps);
  BoolSexpFsm_destroy(module_bool_fsm);
  BoolSexpFsm_destroy(environment_bool_fsm);

  /* assert that the combination of fsm is equivalent to the monolithic fsm */

  #if ISC_DEBUG

  if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
    fprintf(nusmv_stderr, "Verifying if split fsm (module/environment) is "
        "logically equivalent to the original monolithic fsm.\n");
  }

  be_ptr tmp_1 = BeFsm_get_invar(monolithic_be_fsm);
  tmp_1 = BeEnc_untimed_expr_to_timed(be_enc, tmp_1, 0);

  be_ptr tmp_2 = Be_And(be_mgr, BeFsm_get_invar(*module_be_fsm),
                                BeFsm_get_invar(*environment_be_fsm));
  tmp_2 = BeEnc_untimed_expr_to_timed(be_enc, tmp_2, 0);

  if (tmp_1 != tmp_2) {
    /* Invariant constraints of split fsms have a different structure but
     *   "are logically equivalent to the monolithic version. */
    nusmv_assert(are_logically_equivalent(be_enc, be_mgr, tmp_1, tmp_2, 1, trace_symbols));
    if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
      fprintf(nusmv_stderr, "invars are logically equivalent\n");
    }
  }
  else if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
    fprintf(nusmv_stderr, "invars are isomorphic\n");
  }

  tmp_1 = BeFsm_get_init(monolithic_be_fsm);
  tmp_1 = BeEnc_untimed_expr_to_timed(be_enc, tmp_1, 0);

  tmp_2 = Be_And(be_mgr, BeFsm_get_init(*module_be_fsm),
                         BeFsm_get_init(*environment_be_fsm));
  tmp_2 = BeEnc_untimed_expr_to_timed(be_enc, tmp_2, 0);
 
  if (tmp_1 != tmp_2) {
    /* Initial constraints of split fsms have a different structure but
     *   "are logically equivalent to the monolithic version. */
    nusmv_assert(are_logically_equivalent(be_enc, be_mgr, tmp_1, tmp_2, 1, trace_symbols));
    if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
      fprintf(nusmv_stderr, "inits are logically equivalent\n");
    }
  }
  else if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
    fprintf(nusmv_stderr, "inits are isomorphic\n");
  }

  tmp_1 = BeFsm_get_trans(monolithic_be_fsm);
  tmp_1 = BeEnc_untimed_expr_to_timed(be_enc, tmp_1, 0);

  tmp_2 = Be_And(be_mgr, BeFsm_get_trans(*module_be_fsm),
                         BeFsm_get_trans(*environment_be_fsm));
  tmp_2 = BeEnc_untimed_expr_to_timed(be_enc, tmp_2, 0);

  if (tmp_1 != tmp_2) {
    /* Transition constraints of split fsms have a different structure but
     *   "are logically equivalent to the monolithic version. */
    nusmv_assert(are_logically_equivalent(be_enc, be_mgr, tmp_1, tmp_2, 1, trace_symbols));
    if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
      fprintf(nusmv_stderr, "trans are logically equivalent\n");
    }
  }
  else if (opt_verbose_level_gt(OptsHandler_get_instance(), 0)) {
    fprintf(nusmv_stderr, "trans are isomorphic\n");
  }

  #endif
}

/**Function********************************************************************

   Synopsis           [Tests if two formulas are logically equivalent.]

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
static boolean are_logically_equivalent(BeEnc_ptr be_enc, Be_Manager_ptr be_mgr,
    be_ptr A, be_ptr B, int k, NodeList_ptr trace_symbols)
{
  /* assumption: A and B don't contain untimed variables */
  be_ptr iff = Be_Iff(be_mgr, A, B);
  return is_valid(be_enc, be_mgr, iff, k, trace_symbols);
}

/**Function********************************************************************

   Synopsis           [Checks if the problem is a validity, i.e. the negation
                       is unsatisfiable.]

   Description        []

   SideEffects        []

   SeeAlso            []

******************************************************************************/
static boolean is_valid(BeEnc_ptr be_enc, Be_Manager_ptr be_mgr, be_ptr prob,
    int k, NodeList_ptr trace_symbols)
{
  be_ptr neg_prob = Be_Not(be_mgr, prob);

  Trace_ptr trace = TRACE(NULL);
  Trace_ptr* traceptr = -1 == k ? (Trace_ptr*)NULL : &trace;

  SatSolverResult res = isc_core_is_sat(be_enc, be_mgr, neg_prob, k, traceptr, trace_symbols);

  /* print trace? */
  if (opt_verbose_level_gt(OptsHandler_get_instance(), 10)) {
    if ((SAT_SOLVER_SATISFIABLE_PROBLEM == res) && (TRACE(NULL) != trace)) {
        int trace_idx = TraceManager_register_trace(global_trace_manager, trace);
        TraceManager_execute_plugin(global_trace_manager, TRACE_OPT(NULL),
                                    TRACE_MANAGER_DEFAULT_PLUGIN,
                                    trace_idx);
    }
  }

  return (SAT_SOLVER_UNSATISFIABLE_PROBLEM == res);
}

/**Function********************************************************************

   Synopsis           [Checks if a boolean expression is satisfiable.]

   Description        [Computes a counterexample if requested and possible.]

   SideEffects        []

   SeeAlso            []

******************************************************************************/
static SatSolverResult isc_core_is_sat (BeEnc_ptr be_enc, Be_Manager_ptr be_mgr,
    be_ptr beProblem, int k, Trace_ptr* trace, NodeList_ptr trace_symbols)
{
  SatSolverResult sat_res;
  SatSolver_ptr solver;
  Be_Cnf_ptr cnf;

  nusmv_assert((((Trace_ptr*)NULL == trace) && (-1 == k)) ||
               (((Trace_ptr*)NULL != trace) && (-1 < k)));

  solver = Sat_CreateNonIncSolver(get_sat_solver(OptsHandler_get_instance()));  
  if (solver == SAT_SOLVER(NULL)) {
    fprintf(nusmv_stderr, 
	    "Non-incremental sat solver '%s' is not available.\n", 
	    get_sat_solver(OptsHandler_get_instance()));
    
    return 1;
  }

  //printf("apply inlining...\n");
  //beProblem = Bmc_Utils_apply_inlining(be_mgr, beProblem);
  //printf("convert to CNF...\n");
  cnf = Be_ConvertToCnf(be_mgr, beProblem, 1);

  //printf("adding to solver...\n");
  SatSolver_add(solver, cnf, SatSolver_get_permanent_group(solver));
  //printf("set polarity...\n");
  SatSolver_set_polarity(solver, cnf, 1, 
			 SatSolver_get_permanent_group(solver));

  //printf("start solving...\n");
  sat_res = SatSolver_solve_all_groups(solver);

  switch (sat_res) {
  
  case SAT_SOLVER_UNSATISFIABLE_PROBLEM:
    break;

  case SAT_SOLVER_SATISFIABLE_PROBLEM:
    {
      if ((Trace_ptr*)NULL != trace) {
        *trace = Bmc_Utils_generate_cntexample(be_enc, solver,
                                              beProblem,
                                              k, "ISC Counterexample",
                                              trace_symbols);
      }
      break;
    }
  case SAT_SOLVER_INTERNAL_ERROR:
    internal_error("Sorry, solver answered with a fatal Internal "
		   "Failure during problem solving.\n");
    
  case SAT_SOLVER_TIMEOUT:
  case SAT_SOLVER_MEMOUT:
    internal_error("Sorry, solver ran out of resources and aborted "
		   "the execution.\n");

  default:
    internal_error(" Bmc_Simulate: Unexpected value in sat result");
  } /* switch */

  SatSolver_destroy(solver);
  Be_Cnf_Delete(cnf);

  return sat_res;
}

